# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: product_db.proto
# Protobuf Python Version: 6.33.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    1,
    '',
    'product_db.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10product_db.proto\"D\n\x0fLanguageSetting\x12\x10\n\x08language\x18\x01 \x01(\t\x12\x1f\n\x06option\x18\x02 \x01(\x0e\x32\x0f.LanguageOption\"\xdb\x02\n\x0cUserSettings\x12\x14\n\x0cinstall_path\x18\x01 \x01(\t\x12\x13\n\x0bplay_region\x18\x02 \x01(\t\x12)\n\x10\x64\x65sktop_shortcut\x18\x03 \x01(\x0e\x32\x0f.ShortcutOption\x12+\n\x12startmenu_shortcut\x18\x04 \x01(\x0e\x32\x0f.ShortcutOption\x12/\n\x11language_settings\x18\x05 \x01(\x0e\x32\x14.LanguageSettingType\x12\x1e\n\x16selected_text_language\x18\x06 \x01(\t\x12 \n\x18selected_speech_language\x18\x07 \x01(\t\x12#\n\tlanguages\x18\x08 \x03(\x0b\x32\x10.LanguageSetting\x12\x19\n\x11gfx_override_tags\x18\t \x01(\t\x12\x15\n\rversionbranch\x18\n \x01(\t\"Q\n\x10InstallHandshake\x12\x0f\n\x07product\x18\x01 \x01(\t\x12\x0b\n\x03uid\x18\x02 \x01(\t\x12\x1f\n\x08settings\x18\x03 \x01(\x0b\x32\r.UserSettings\"3\n\x0b\x42uildConfig\x12\x0e\n\x06region\x18\x01 \x01(\t\x12\x14\n\x0c\x62uild_config\x18\x02 \x01(\t\"\xf4\x02\n\x10\x42\x61seProductState\x12\x11\n\tinstalled\x18\x01 \x01(\x08\x12\x10\n\x08playable\x18\x02 \x01(\x08\x12\x17\n\x0fupdate_complete\x18\x03 \x01(\x08\x12%\n\x1d\x62\x61\x63kground_download_available\x18\x04 \x01(\x08\x12$\n\x1c\x62\x61\x63kground_download_complete\x18\x05 \x01(\x08\x12\x17\n\x0f\x63urrent_version\x18\x06 \x01(\t\x12\x1b\n\x13\x63urrent_version_str\x18\x07 \x01(\t\x12,\n\x16installed_build_config\x18\x08 \x03(\x0b\x32\x0c.BuildConfig\x12\x36\n background_download_build_config\x18\t \x03(\x0b\x32\x0c.BuildConfig\x12\x16\n\x0e\x64\x65\x63ryption_key\x18\n \x01(\t\x12!\n\x19\x63ompleted_install_actions\x18\x0b \x03(\t\"h\n\x10\x42\x61\x63kfillProgress\x12\x10\n\x08progress\x18\x01 \x01(\x01\x12\x1a\n\x12\x62\x61\x63kgrounddownload\x18\x02 \x01(\x08\x12\x0e\n\x06paused\x18\x03 \x01(\x08\x12\x16\n\x0e\x64ownload_limit\x18\x04 \x01(\x04\"\"\n\x0eRepairProgress\x12\x10\n\x08progress\x18\x01 \x01(\x01\"\x8b\x01\n\x0eUpdateProgress\x12\x1a\n\x12last_disc_set_used\x18\x01 \x01(\t\x12\x10\n\x08progress\x18\x02 \x01(\x01\x12\x14\n\x0c\x64isc_ignored\x18\x03 \x01(\x08\x12\x19\n\x11total_to_download\x18\x04 \x01(\x04\x12\x1a\n\x12\x64ownload_remaining\x18\x05 \x01(\x04\"\xc5\x01\n\x12\x43\x61\x63hedProductState\x12-\n\x12\x62\x61se_product_state\x18\x01 \x01(\x0b\x32\x11.BaseProductState\x12,\n\x11\x62\x61\x63kfill_progress\x18\x02 \x01(\x0b\x32\x11.BackfillProgress\x12(\n\x0frepair_progress\x18\x03 \x01(\x0b\x32\x0f.RepairProgress\x12(\n\x0fupdate_progress\x18\x04 \x01(\x0b\x32\x0f.UpdateProgress\"K\n\x11ProductOperations\x12$\n\x10\x61\x63tive_operation\x18\x01 \x01(\x0e\x32\n.Operation\x12\x10\n\x08priority\x18\x02 \x01(\x04\"\xb7\x01\n\x0eProductInstall\x12\x0b\n\x03uid\x18\x01 \x01(\t\x12\x14\n\x0cproduct_code\x18\x02 \x01(\t\x12\x1f\n\x08settings\x18\x03 \x01(\x0b\x32\r.UserSettings\x12\x31\n\x14\x63\x61\x63hed_product_state\x18\x04 \x01(\x0b\x32\x13.CachedProductState\x12.\n\x12product_operations\x18\x05 \x01(\x0b\x32\x12.ProductOperations\"O\n\rProductConfig\x12\x14\n\x0cproduct_code\x18\x01 \x01(\t\x12\x15\n\rmetadata_hash\x18\x02 \x01(\t\x12\x11\n\ttimestamp\x18\x03 \x01(\t\"?\n\rActiveProcess\x12\x14\n\x0cprocess_name\x18\x01 \x01(\t\x12\x0b\n\x03pid\x18\x02 \x01(\x05\x12\x0b\n\x03uri\x18\x03 \x03(\t\"B\n\x10\x44ownloadSettings\x12\x16\n\x0e\x64ownload_limit\x18\x01 \x01(\x05\x12\x16\n\x0e\x62\x61\x63kfill_limit\x18\x02 \x01(\x05\"\xe3\x01\n\tProductDb\x12)\n\x10product_installs\x18\x01 \x03(\x0b\x32\x0f.ProductInstall\x12*\n\x0f\x61\x63tive_installs\x18\x02 \x03(\x0b\x32\x11.InstallHandshake\x12(\n\x10\x61\x63tive_processes\x18\x03 \x03(\x0b\x32\x0e.ActiveProcess\x12\'\n\x0fproduct_configs\x18\x04 \x03(\x0b\x32\x0e.ProductConfig\x12,\n\x11\x64ownload_settings\x18\x05 \x01(\x0b\x32\x11.DownloadSettings*q\n\x0eLanguageOption\x12\x13\n\x0fLANGOPTION_NONE\x10\x00\x12\x13\n\x0fLANGOPTION_TEXT\x10\x01\x12\x15\n\x11LANGOPTION_SPEECH\x10\x02\x12\x1e\n\x1aLANGOPTION_TEXT_AND_SPEECH\x10\x03*u\n\x13LanguageSettingType\x12\x14\n\x10LANGSETTING_NONE\x10\x00\x12\x16\n\x12LANGSETTING_SINGLE\x10\x01\x12\x16\n\x12LANGSETTING_SIMPLE\x10\x02\x12\x18\n\x14LANGSETTING_ADVANCED\x10\x03*N\n\x0eShortcutOption\x12\x11\n\rSHORTCUT_NONE\x10\x00\x12\x11\n\rSHORTCUT_USER\x10\x01\x12\x16\n\x12SHORTCUT_ALL_USERS\x10\x02*P\n\tOperation\x12\r\n\tOP_UPDATE\x10\x00\x12\x0f\n\x0bOP_BACKFILL\x10\x01\x12\r\n\tOP_REPAIR\x10\x02\x12\x14\n\x07OP_NONE\x10\xff\xff\xff\xff\xff\xff\xff\xff\xff\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'product_db_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_LANGUAGEOPTION']._serialized_start=2142
  _globals['_LANGUAGEOPTION']._serialized_end=2255
  _globals['_LANGUAGESETTINGTYPE']._serialized_start=2257
  _globals['_LANGUAGESETTINGTYPE']._serialized_end=2374
  _globals['_SHORTCUTOPTION']._serialized_start=2376
  _globals['_SHORTCUTOPTION']._serialized_end=2454
  _globals['_OPERATION']._serialized_start=2456
  _globals['_OPERATION']._serialized_end=2536
  _globals['_LANGUAGESETTING']._serialized_start=20
  _globals['_LANGUAGESETTING']._serialized_end=88
  _globals['_USERSETTINGS']._serialized_start=91
  _globals['_USERSETTINGS']._serialized_end=438
  _globals['_INSTALLHANDSHAKE']._serialized_start=440
  _globals['_INSTALLHANDSHAKE']._serialized_end=521
  _globals['_BUILDCONFIG']._serialized_start=523
  _globals['_BUILDCONFIG']._serialized_end=574
  _globals['_BASEPRODUCTSTATE']._serialized_start=577
  _globals['_BASEPRODUCTSTATE']._serialized_end=949
  _globals['_BACKFILLPROGRESS']._serialized_start=951
  _globals['_BACKFILLPROGRESS']._serialized_end=1055
  _globals['_REPAIRPROGRESS']._serialized_start=1057
  _globals['_REPAIRPROGRESS']._serialized_end=1091
  _globals['_UPDATEPROGRESS']._serialized_start=1094
  _globals['_UPDATEPROGRESS']._serialized_end=1233
  _globals['_CACHEDPRODUCTSTATE']._serialized_start=1236
  _globals['_CACHEDPRODUCTSTATE']._serialized_end=1433
  _globals['_PRODUCTOPERATIONS']._serialized_start=1435
  _globals['_PRODUCTOPERATIONS']._serialized_end=1510
  _globals['_PRODUCTINSTALL']._serialized_start=1513
  _globals['_PRODUCTINSTALL']._serialized_end=1696
  _globals['_PRODUCTCONFIG']._serialized_start=1698
  _globals['_PRODUCTCONFIG']._serialized_end=1777
  _globals['_ACTIVEPROCESS']._serialized_start=1779
  _globals['_ACTIVEPROCESS']._serialized_end=1842
  _globals['_DOWNLOADSETTINGS']._serialized_start=1844
  _globals['_DOWNLOADSETTINGS']._serialized_end=1910
  _globals['_PRODUCTDB']._serialized_start=1913
  _globals['_PRODUCTDB']._serialized_end=2140
# @@protoc_insertion_point(module_scope)



# Standard Library
import os

# Lutris Modules
from lutris.settings import RUNTIME_DIR
from lutris.util import system
from lutris.util.log import logger


class ControllerMapping:
    valid_keys = [
        "platform",
        "leftx",
        "lefty",
        "rightx",
        "righty",
        "a",
        "b",
        "back",
        "dpdown",
        "dpleft",
        "dpright",
        "dpup",
        "guide",
        "leftshoulder",
        "leftstick",
        "lefttrigger",
        "rightshoulder",
        "rightstick",
        "righttrigger",
        "start",
        "x",
        "y",
    ]

    def __init__(self, guid, name, mapping):
        self.guid = guid
        self.name = name
        self.mapping = mapping
        self.keys = {}
        self.parse()

    def __str__(self):
        return self.name

    def parse(self):
        key_maps = self.mapping.split(",")
        for key_map in key_maps:
            if not key_map:
                continue
            xinput_key, sdl_key = key_map.split(":")
            if xinput_key not in self.valid_keys:
                logger.warning("Unrecognized key %s", xinput_key)
                continue
            self.keys[xinput_key] = sdl_key


class GameControllerDB:
    db_path = os.path.join(RUNTIME_DIR, "gamecontrollerdb/gamecontrollerdb.txt")

    def __init__(self):
        if not system.path_exists(self.db_path):
            raise OSError("Path to gamecontrollerdb.txt not provided or invalid")
        self.controllers = {}
        self.parsedb()

    def __str__(self):
        return "GameControllerDB <%s>" % self.db_path

    def __getitem__(self, value):
        return self.controllers[value]

    def parsedb(self):
        with open(self.db_path, "r", encoding="utf-8") as db:
            for line in db.readlines():
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                guid, name, mapping = line.strip().split(",", 2)
                self.controllers[guid] = ControllerMapping(guid, name, mapping)



"""THIS MODULE IS UNMAINTAINTED AND WILL BE MARKED FOR DEPRECATION UNLESS SOMEONE VOLUNTEERS TO PROPERLY MAINTAIN IT."""

from pypresence import DiscordNotFound, Presence

from lutris.util.discord.base import DiscordRichPresenceBase


class DiscordRichPresenceClient(DiscordRichPresenceBase):
    rpc = None  # Presence Object

    def __init__(self):
        self.playing = None
        self.rpc = None

    def update(self, discord_id):
        if self.rpc is not None:
            # Clear the old RPC before creating a new one
            self.clear()

        # Create a new Presence object with the desired app id
        self.rpc = Presence(discord_id)
        try:
            self.rpc.connect()
        except DiscordNotFound:
            return

        self.rpc.update()

    def clear(self):
        if self.rpc is None:
            # Skip already deleted rpc
            return
        # Clear and Close Presence connection
        self.rpc.clear()
        self.rpc.close()
        # Clear Presence Object
        self.rpc = None
        # Clear Internal Reference
        self.playing = None



import os

from lutris.util.log import logger
from lutris.util.ubisoft.consts import UBISOFT_REGISTRY_LAUNCHER_INSTALLS
from lutris.util.wine.registry import WineRegistry

NOT_INSTALLED = "NOT_INSTALLED"
INSTALLED = "INSTALLED"


def _get_registry_value_from_path(registry_path, key):
    registry = WineRegistry()
    return registry.query(registry_path, key)


def _return_local_game_path_from_special_registry(special_registry_path):
    if not special_registry_path:
        return NOT_INSTALLED
    return _get_registry_value_from_path("HKEY_LOCAL_MACHINE" + special_registry_path, "InstallLocation")


def _return_local_game_path(launch_id):
    installs_path = UBISOFT_REGISTRY_LAUNCHER_INSTALLS
    registry = WineRegistry()
    game_path = registry.query("HKEY_LOCAL_MACHINE" + installs_path + f"\\{launch_id}", "InstallDir")
    return os.path.normcase(os.path.normpath(game_path))


def get_local_game_path(special_registry_path, launch_id):
    local_game_path = _return_local_game_path(launch_id)
    if not local_game_path and special_registry_path:
        local_game_path = _return_local_game_path_from_special_registry(special_registry_path)
    return local_game_path


async def get_size_at_path(start_path):
    total_size = 0
    for dirpath, _dirnames, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)
    return total_size


def _is_file_at_path(path, file):
    if os.path.isdir(path):
        return os.path.isfile(os.path.join(path, file))
    return False


def _read_status_from_state_file(game_path):
    try:
        if os.path.exists(os.path.join(game_path, "uplay_install.state")):
            with open(os.path.join(game_path, "uplay_install.state"), "rb") as f:
                if f.read()[0] == 0x0A:
                    return INSTALLED
                return NOT_INSTALLED
        return NOT_INSTALLED
    except Exception as e:
        logger.warning("Issue reading install state file for %s: %s", game_path, repr(e))
        return NOT_INSTALLED


def get_game_installed_status(path, exe=None, special_registry_path=None):
    status = NOT_INSTALLED
    try:
        if path and os.access(path, os.F_OK):
            status = _read_status_from_state_file(path)
            # Fallback for old games
            if status == NOT_INSTALLED and exe and special_registry_path:
                if _is_file_at_path(path, exe):
                    status = INSTALLED
    except Exception as e:
        logger.error("Error reading game installed status at %s, %s", path, repr(e))
    return status


def get_ubisoft_registry(prefix, fullpath):
    """Get a value from the registry in a Ubisoft Connect prefix"""
    if not fullpath:
        return ""
    if fullpath.startswith("HKEY"):
        path, key = fullpath.rsplit("\\", maxsplit=1)
        path = convert_ubisoft_key(path)
        return prefix.get_registry_key(path, key)
    return fullpath


def convert_ubisoft_key(key_path):
    """Convert Ubisoft registry keys for Wine compatibility"""
    if "LOCAL_MACHINE" in key_path:
        key_path = key_path.replace("SOFTWARE\\", "Software\\Wow6432Node\\")
    key_path = key_path.replace("\\UBISOFT", "\\Ubisoft")
    key_path = key_path.replace("\\", "/")
    return key_path



import gzip
import os
import shutil
import subprocess
import tarfile
import uuid
import zlib
from typing import List, Tuple

from lutris import settings
from lutris.exceptions import MissingExecutableError
from lutris.util import system
from lutris.util.log import logger


class ExtractError(Exception):
    """Exception raised when and archive fails to extract"""


def extract_archive(path: str, to_directory: str = ".", merge_single: bool = True, extractor=None) -> Tuple[str, str]:
    path = os.path.abspath(path)
    logger.debug("Extracting %s to %s", path, to_directory)

    if extractor is None:
        extractor = _guess_extractor(path)

    opener, mode = _get_archive_opener(extractor)

    temp_path = temp_dir = os.path.join(to_directory, ".extract-%s" % _random_id())
    try:
        _do_extract(path, temp_path, opener, mode, extractor)
    except (OSError, zlib.error, tarfile.ReadError, EOFError) as ex:
        logger.error("Extraction failed: %s", ex)
        raise ExtractError(str(ex)) from ex
    if merge_single:
        extracted = os.listdir(temp_path)
        if len(extracted) == 1:
            temp_path = os.path.join(temp_path, extracted[0])

    if os.path.isfile(temp_path):
        destination_path = os.path.join(to_directory, extracted[0])
        if os.path.isfile(destination_path):
            logger.warning("Overwrite existing file %s", destination_path)
            os.remove(destination_path)
        if os.path.isdir(destination_path):
            os.rename(destination_path, destination_path + _random_id())

        shutil.move(temp_path, to_directory)
        os.removedirs(temp_dir)
    else:
        for archive_file in os.listdir(temp_path):
            source_path = os.path.join(temp_path, archive_file)
            destination_path = os.path.join(to_directory, archive_file)
            # logger.debug("Moving extracted files from %s to %s", source_path, destination_path)

            if system.path_exists(destination_path):
                logger.warning("Overwrite existing path %s", destination_path)
                if os.path.isfile(destination_path):
                    os.remove(destination_path)
                    shutil.move(source_path, destination_path)
                elif os.path.isdir(destination_path):
                    try:
                        system.merge_folders(source_path, destination_path)
                    except OSError as ex:
                        logger.error(
                            "Failed to merge to destination %s: %s",
                            destination_path,
                            ex,
                        )
                        raise ExtractError(str(ex)) from ex
            else:
                shutil.move(source_path, destination_path)
        system.delete_folder(temp_dir)
    logger.debug("Finished extracting %s to %s", path, to_directory)
    return path, to_directory


def _guess_extractor(path):
    """Guess what extractor should be used from a file name"""
    if path.endswith(".tar"):
        extractor = "tar"
    elif path.endswith((".tar.gz", ".tgz")):
        extractor = "tgz"
    elif path.endswith((".tar.xz", ".txz", ".tar.lzma")):
        extractor = "txz"
    elif path.endswith((".tar.bz2", ".tbz2", ".tbz")):
        extractor = "tbz2"
    elif path.endswith((".tar.zst", ".tzst")):
        extractor = "tzst"
    elif path.endswith(".gz"):
        extractor = "gzip"
    elif path.endswith(".exe"):
        extractor = "exe"
    elif path.endswith(".deb"):
        extractor = "deb"
    elif path.casefold().endswith(".appimage"):
        extractor = "AppImage"
    else:
        extractor = None
    return extractor


def _get_archive_opener(extractor):
    """Return the archive opener and optional mode for an extractor"""
    mode = None
    if extractor == "tar":
        opener, mode = tarfile.open, "r:"
    elif extractor == "tgz":
        opener, mode = tarfile.open, "r:gz"
    elif extractor == "txz":
        opener, mode = tarfile.open, "r:xz"
    elif extractor in ("tbz2", "bz2"):  # bz2 is used for .tar.bz2 in some installer scripts
        opener, mode = tarfile.open, "r:bz2"
    elif extractor == "tzst":
        opener, mode = tarfile.open, "r:zst"  # Note: not supported by tarfile yet
    elif extractor == "gzip":
        opener = "gz"
    elif extractor == "gog":
        opener = "innoextract"
    elif extractor == "exe":
        opener = "exe"
    elif extractor == "deb":
        opener = "deb"
    elif extractor == "AppImage":
        opener = "AppImage"
    else:
        opener = "7zip"
    return opener, mode


def _random_id():
    """Return a random ID"""
    return str(uuid.uuid4())[:8]


def _do_extract(archive: str, dest: str, opener, mode: str = None, extractor=None) -> None:
    if opener == "gz":
        _decompress_gz(archive, dest)
    elif opener == "7zip":
        _extract_7zip(archive, dest, archive_type=extractor)
    elif opener == "exe":
        _extract_exe(archive, dest)
    elif opener == "innoextract":
        _extract_gog(archive, dest)
    elif opener == "deb":
        _extract_deb(archive, dest)
    elif opener == "AppImage":
        _extract_AppImage(archive, dest)
    else:
        handler = opener(archive, mode)
        handler.extractall(dest)
        handler.close()


def _decompress_gz(file_path: str, dest_path: str):
    """Decompress a gzip file."""
    if dest_path:
        dest_filename = os.path.join(dest_path, os.path.basename(file_path[:-3]))
    else:
        dest_filename = file_path[:-3]
    os.makedirs(os.path.dirname(dest_filename), exist_ok=True)

    with open(dest_filename, "wb") as dest_file:
        gzipped_file = gzip.open(file_path, "rb")
        dest_file.write(gzipped_file.read())
        gzipped_file.close()


def _extract_7zip(path: str, dest: str, archive_type: str = None) -> None:
    _7zip_path = os.path.join(settings.RUNTIME_DIR, "p7zip/7z")
    if not system.path_exists(_7zip_path):
        _7zip_path = system.find_required_executable("7z")
    command = [_7zip_path, "x", path, "-o{}".format(dest), "-aoa"]
    if archive_type and archive_type != "auto":
        command.append("-t{}".format(archive_type))
    subprocess.call(command)


def _extract_exe(path: str, dest: str) -> None:
    if _check_inno_exe(path):
        _decompress_gog(path, dest)
    else:
        # use 7za to check if exe is an archive
        _7zip_path = os.path.join(settings.RUNTIME_DIR, "p7zip/7za")
        if not system.path_exists(_7zip_path):
            _7zip_path = system.find_required_executable("7za")
        command = [_7zip_path, "t", path]
        return_code = subprocess.call(command)
        if return_code == 0:
            _extract_7zip(path, dest)
        else:
            raise RuntimeError("specified exe is not an archive or GOG setup file")


def _check_inno_exe(path) -> bool:
    """Check if a path in a compatible innosetup archive"""
    try:
        innoextract_path = _get_innoextract_path()
    except MissingExecutableError:
        logger.warning("Innoextract not found, can't determine type of archive %s", path)
        return False

    command = [innoextract_path, "-i", path]
    return_code = subprocess.call(command)
    return return_code == 0


def _decompress_gog(file_path: str, destination_path: str) -> None:
    innoextract_path = _get_innoextract_path()
    system.create_folder(destination_path)  # innoextract cannot do mkdir -p
    process = subprocess.run(
        [innoextract_path, "-m", "-g", "-d", destination_path, "-e", file_path], capture_output=True
    )
    if process.returncode != 0:
        stderr = process.stderr.decode().strip()
        logger.error("Innoextract failed to extract GOG setup file: %s", stderr)
        raise RuntimeError("innoextract failed to extract GOG setup file")


def _get_innoextract_path() -> str:
    """Return the path where innoextract is installed"""
    inno_dirs = [path for path in os.listdir(settings.RUNTIME_DIR) if path.startswith("innoextract")]
    for inno_dir in inno_dirs:
        inno_path = os.path.join(settings.RUNTIME_DIR, inno_dir, "innoextract")
        if system.path_exists(inno_path):
            return inno_path

    inno_path = system.find_required_executable("innoextract")
    logger.warning("innoextract not available in the runtime folder, using some random version")
    return inno_path


def _extract_gog(path: str, dest: str) -> None:
    if _check_inno_exe(path):
        _decompress_gog(path, dest)
    else:
        raise RuntimeError("specified exe is not a GOG setup file")


def _extract_deb(archive: str, dest: str) -> None:
    """Extract the contents of a deb file to a destination folder"""
    _extract_7zip(archive, dest, archive_type="ar")
    debian_folder = os.path.join(dest, "debian")
    os.makedirs(debian_folder)

    control_file_exts = [".gz", ".xz", ".zst", ""]
    for extension in control_file_exts:
        control_tar_path = os.path.join(dest, "control.tar{}".format(extension))
        if os.path.exists(control_tar_path):
            shutil.move(control_tar_path, debian_folder)
            break

    data_file_exts = [".gz", ".xz", ".zst", ".bz2", ".lzma", ""]
    for extension in data_file_exts:
        data_tar_path = os.path.join(dest, "data.tar{}".format(extension))
        if os.path.exists(data_tar_path):
            extract_archive(data_tar_path, dest)
            os.remove(data_tar_path)
            break


def _extract_AppImage(path: str, dest: str) -> None:
    """This is really here to prevent 7-zip from extracting the AppImage;
    we want to just use this sort of file as-is."""
    system.create_folder(dest)
    shutil.copy(path, dest)


def get_innoextract_list(file_path: str) -> List[str]:
    """Return the list of files contained in a GOG archive"""
    output = system.read_process_output([_get_innoextract_path(), "-lmq", file_path])
    return [line[3:] for line in output.split("\n") if line]



"""Interact with an exiting EGS install"""

import json
import os

from lutris.util.log import logger


class EGSLauncher:
    manifests_paths = "ProgramData/Epic/EpicGamesLauncher/Data/Manifests"

    def __init__(self, prefix_path):
        self.prefix_path = prefix_path

    def iter_manifests(self):
        manifests_path = os.path.join(self.prefix_path, "drive_c", self.manifests_paths)
        if not os.path.exists(manifests_path):
            logger.warning("No valid path for EGS games manifests in %s", manifests_path)
            return []
        for manifest in os.listdir(manifests_path):
            if not manifest.endswith(".item"):
                continue
            with open(os.path.join(manifests_path, manifest), encoding="utf-8") as manifest_file:
                manifest_content = json.loads(manifest_file.read())
            if manifest_content["MainGameAppName"] != manifest_content["AppName"]:
                continue
            yield manifest_content



from gettext import gettext as _

from lutris.gui.config.game_common import GameDialogCommon


class EditGameConfigDialog(GameDialogCommon):
    """Game config edit dialog."""

    def __init__(self, parent, game):
        super().__init__(_("Configure %s") % game.name, config_level="game", parent=parent)
        self.game = game
        self.lutris_config = game.config
        self.runner_name = game.runner_name
        self.build_notebook()
        self.build_tabs()
        self.show_all()



# Standard Library
import os
import shlex
from gettext import gettext as _

from lutris import settings
from lutris.exceptions import MissingGameExecutableError

# Lutris Modules
from lutris.runners.commands.dosbox import dosexec, makeconfig  # NOQA pylint: disable=unused-import
from lutris.runners.runner import Runner
from lutris.util import system


class dosbox(Runner):
    human_name = _("DOSBox")
    description = _("MS-DOS emulator")
    platforms = [_("MS-DOS")]
    runnable_alone = True
    runner_executable = "dosbox/dosbox"
    flatpak_id = "io.github.dosbox-staging"
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("Main file"),
            "help": _(
                "The CONF, EXE, COM or BAT file to launch.\n"
                "If the executable is managed in the config file, this should be the config file, "
                "instead specifying it in 'Configuration file'."
            ),
        },
        {
            "option": "config_file",
            "type": "file",
            "label": _("Configuration file"),
            "help": _(
                "Start DOSBox with the options specified in this file. \n"
                "It can have a section in which you can put commands "
                "to execute on startup. Read DOSBox's documentation "
                "for more information."
            ),
        },
        {
            "option": "args",
            "type": "string",
            "label": _("Command line arguments"),
            "help": _("Command line arguments used when launching DOSBox"),
            "validator": shlex.split,
        },
        {
            "option": "working_dir",
            "type": "directory",
            "label": _("Working directory"),
            "warn_if_non_writable_parent": True,
            "help": _(
                "The location where the game is run from.\nBy default, Lutris uses the directory of the executable."
            ),
        },
    ]

    runner_options = [
        {
            "option": "fullscreen",
            "section": _("Graphics"),
            "label": _("Open game in fullscreen"),
            "type": "bool",
            "default": False,
            "help": _("Tells DOSBox to launch the game in fullscreen."),
        },
        {
            "option": "exit",
            "label": _("Exit DOSBox with the game"),
            "type": "bool",
            "default": True,
            "help": _("Shut down DOSBox when the game is quit."),
        },
    ]

    def make_absolute(self, path):
        """Return a guaranteed absolute path"""
        if not path:
            return ""
        path = os.path.expanduser(path)
        if os.path.isabs(path):
            return path
        directory = self.game_data.get("directory")
        if directory:
            directory = os.path.expanduser(directory)
            return os.path.join(directory, path)
        return ""

    @property
    def main_file(self):
        return self.make_absolute(self.game_config.get("main_file"))

    @property
    def libs_dir(self):
        path = os.path.join(settings.RUNNER_DIR, "dosbox/lib")
        return path if system.path_exists(path) else ""

    def get_run_data(self):
        env = self.get_env()
        env["LD_LIBRARY_PATH"] = os.pathsep.join(filter(None, [self.libs_dir, env.get("LD_LIBRARY_PATH")]))
        return {"env": env, "command": self.get_command()}

    @property
    def working_dir(self):
        """Return the working directory to use when running the game."""
        option = self.game_config.get("working_dir")
        if option:
            return os.path.expanduser(option)
        if self.main_file:
            return os.path.dirname(self.main_file)
        return super().working_dir

    def play(self):
        main_file = self.main_file
        if not system.path_exists(main_file):
            raise MissingGameExecutableError(filename=main_file)
        args = shlex.split(self.game_config.get("args") or "")
        command = self.get_command()

        if main_file.endswith(".conf"):
            command.append("-conf")
            command.append(main_file)
        else:
            command.append(main_file)
        # Options
        if self.game_config.get("config_file"):
            command.append("-conf")
            command.append(self.make_absolute(self.game_config["config_file"]))

        if self.runner_config.get("fullscreen"):
            command.append("-fullscreen")

        if self.runner_config.get("exit"):
            command.append("-exit")

        if args:
            command.extend(args)

        return {"command": command, "ld_library_path": self.libs_dir}



from datetime import datetime
from gettext import gettext as _

from gi.repository import Gtk, Pango

from lutris import runners, services
from lutris.database.games import get_game_for_service
from lutris.game import GAME_INSTALLED, GAME_START, GAME_STARTED, GAME_STOPPED, GAME_UPDATED, Game
from lutris.game_actions import get_game_actions
from lutris.gui.widgets.contextual_menu import update_action_widget_visibility
from lutris.util.strings import gtk_safe


class GameBar(Gtk.Box):
    def __init__(self, db_game, application, window):
        """Create the game bar with a database row"""
        super().__init__(
            orientation=Gtk.Orientation.VERTICAL,
            visible=True,
            margin_top=12,
            margin_left=12,
            margin_bottom=12,
            margin_right=12,
            spacing=6,
        )

        self.application = application
        self.window = window

        self.game_start_registration = GAME_START.register(self.on_game_state_changed)
        self.game_started_registration = GAME_STARTED.register(self.on_game_state_changed)
        self.game_stopped_registration = GAME_STOPPED.register(self.on_game_state_changed)
        self.game_updated_registration = GAME_UPDATED.register(self.on_game_state_changed)
        self.game_installed_registration = GAME_INSTALLED.register(self.on_game_state_changed)
        self.connect("destroy", self.on_destroy)

        self.set_margin_bottom(12)
        self.db_game = db_game
        self.service = None
        if db_game.get("service"):
            try:
                self.service = services.SERVICES[db_game["service"]]()
            except KeyError:
                pass

        game_id = None
        if "service_id" in db_game:
            self.appid = db_game["service_id"]
            game_id = db_game["id"]
        elif self.service:
            self.appid = db_game["appid"]
            game = get_game_for_service(self.service.id, self.appid)
            if game:
                game_id = game["id"]
        if game_id:
            self.game = application.get_game_by_id(game_id)
        else:
            self.game = Game.create_empty_service_game(db_game, self.service)
        self.update_view()

    def on_destroy(self, widget):
        self.game_start_registration.unregister()
        self.game_started_registration.unregister()
        self.game_stopped_registration.unregister()
        self.game_updated_registration.unregister()
        self.game_installed_registration.unregister()
        return True

    def update_view(self):
        """Populate the view with widgets"""
        game_actions = get_game_actions([self.game], window=self.window, application=self.application)

        game_label = self.get_game_name_label()
        game_label.set_halign(Gtk.Align.START)
        self.pack_start(game_label, False, False, 0)

        hbox = Gtk.Box(Gtk.Orientation.HORIZONTAL, spacing=6)
        self.pack_start(hbox, False, False, 0)

        self.play_button = self.get_play_button(game_actions)
        hbox.pack_start(self.play_button, False, False, 0)

        hbox.pack_start(self.get_runner_button(), False, False, 0)
        hbox.pack_start(self.get_platform_label(), False, False, 0)
        if self.game.lastplayed:
            hbox.pack_start(self.get_last_played_label(), False, False, 0)
        if self.game.playtime:
            hbox.pack_start(self.get_playtime_label(), False, False, 0)
        hbox.show_all()

    @staticmethod
    def get_popover_box(primary_button, popover_buttons, primary_opens_popover=False):
        """Creates a box that contains a primary button and a second button that
        opens a popover with the popover_buttons in it; these have a linked
        style so this looks like a single button.

        If primary_opens_popover is true, this method also handled the 'clicked' signal
        of the primary button to trigger the popover as well."""

        def get_popover(parent):
            # Creates the popover widget containing the list of link buttons
            pop = Gtk.Popover()
            vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, visible=True)
            vbox.set_border_width(9)
            vbox.set_spacing(3)

            for button in popover_buttons:
                vbox.pack_end(button, False, False, 0)

            pop.add(vbox)
            pop.set_position(Gtk.PositionType.TOP)
            pop.set_constrain_to(Gtk.PopoverConstraint.NONE)
            pop.set_relative_to(parent)
            return pop

        box = Gtk.HBox(visible=True)
        style_context = box.get_style_context()
        style_context.add_class("linked")

        box.pack_start(primary_button, False, False, 0)

        if popover_buttons:
            popover_button = Gtk.MenuButton(direction=Gtk.ArrowType.UP, visible=True)
            popover_button.set_size_request(32, 32)
            popover = get_popover(popover_button)
            popover_button.set_popover(popover)
            box.pack_start(popover_button, False, False, 0)

            if primary_opens_popover:
                primary_button.connect("clicked", lambda _x: popover_button.emit("clicked"))

        return box

    def get_link_button(self, text, callback=None):
        """Return a suitable button for a menu popover; this must be
        a ModelButton to be styled correctly."""
        if text == "-":
            return Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL, visible=True)

        button = Gtk.ModelButton(text, visible=True, xalign=0.0)
        if callback:
            button.connect("clicked", self.on_link_button_clicked, callback)
        return button

    def get_game_name_label(self):
        """Return the label with the game's title"""
        title_label = Gtk.Label(visible=True)
        title_label.set_ellipsize(Pango.EllipsizeMode.END)
        title_label.set_markup("<span font_desc='16'><b>%s</b></span>" % gtk_safe(self.game.name))
        return title_label

    def get_runner_button(self):
        if not self.game.has_runner:
            return Gtk.Box()
        icon_name = self.game.runner.name + "-symbolic"
        runner_icon = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.MENU)
        runner_popover_buttons = self.get_runner_buttons()
        if runner_popover_buttons:
            runner_button = Gtk.Button(image=runner_icon, visible=True)
            return GameBar.get_popover_box(runner_button, runner_popover_buttons, primary_opens_popover=True)

        runner_icon.set_margin_left(49)
        runner_icon.set_margin_right(6)
        return runner_icon

    def get_platform_label(self):
        platform_label = Gtk.Label(visible=True)
        if not self.game.platform:
            return platform_label
        platform_label.set_size_request(120, -1)
        platform_label.set_alignment(0, 0.5)
        platform = gtk_safe(self.game.platform)
        platform_label.set_tooltip_markup(platform)
        platform_label.set_markup(_("Platform:\n<b>%s</b>") % platform)
        platform_label.set_property("ellipsize", Pango.EllipsizeMode.END)
        return platform_label

    def get_playtime_label(self):
        """Return the label containing the playtime info"""
        playtime_label = Gtk.Label(visible=True)
        playtime_label.set_size_request(120, -1)
        playtime_label.set_alignment(0, 0.5)
        playtime_label.set_markup(_("Time played:\n<b>%s</b>") % self.game.formatted_playtime)
        return playtime_label

    def get_last_played_label(self):
        """Return the label containing the last played info"""
        last_played_label = Gtk.Label(visible=True)
        last_played_label.set_size_request(120, -1)
        last_played_label.set_alignment(0, 0.5)
        lastplayed = datetime.fromtimestamp(self.game.lastplayed)
        last_played_label.set_markup(_("Last played:\n<b>%s</b>") % lastplayed.strftime("%b %-d %Y"))
        return last_played_label

    def get_locate_installed_game_button(self, game_actions):
        """Return a button to locate an existing install"""
        button = self.get_link_button(_("Locate installed game"))
        button.connect("clicked", game_actions.on_locate_installed_game)
        return button

    def get_play_button(self, game_actions):
        """Return the widget for install/play/stop and game config"""
        button = Gtk.Button(visible=True)
        button.set_size_request(120, 32)
        game_buttons = None

        if self.game.is_installed:
            game_buttons = self.get_game_buttons(game_actions)
            if self.game.state == self.game.STATE_STOPPED:
                button.set_label(_("Play"))
                button.connect("clicked", game_actions.on_game_launch)
                button.set_sensitive(game_actions.is_game_launchable)
            elif self.game.state == self.game.STATE_LAUNCHING:
                button.set_label(_("Launching"))
                button.set_sensitive(False)
            else:
                button.set_label(_("Stop"))
                button.connect("clicked", game_actions.on_game_stop)
                button.set_sensitive(game_actions.is_game_running)
        else:
            button.set_label(_("Install"))
            button.connect("clicked", game_actions.on_install_clicked)
            button.set_sensitive(game_actions.is_installable)
            if self.service:
                if self.service.local:
                    # Local services don't show an install dialog, they can be launched directly
                    button.set_label(_("Play"))
                if self.service.drm_free:
                    game_buttons = [self.get_locate_installed_game_button(game_actions)]

        if game_buttons:
            button.set_size_request(84, 32)
            box = GameBar.get_popover_box(button, game_buttons)
            return box
        return button

    def get_game_buttons(self, game_actions):
        """Return a list of buttons to use in the panel"""
        displayed = game_actions.get_displayed_entries()
        buttons = []
        button_visibility = {}
        for action_id, label, callback in game_actions.get_game_actions():
            if action_id in ("play", "stop", "install"):
                continue
            button = self.get_link_button(label, callback)
            if action_id:
                button_visibility[button] = displayed.get(action_id, True)
            buttons.append(button)

        update_action_widget_visibility(buttons, lambda w: button_visibility.get(w, None))
        return buttons

    def get_runner_buttons(self):
        buttons = []
        if self.game.has_runner and self.game.is_installed:
            runner = runners.import_runner(self.game.runner_name)(self.game.config)
            for _name, label, callback in runner.context_menu_entries:
                button = self.get_link_button(label, callback)
                buttons.append(button)
        return buttons

    def on_link_button_clicked(self, button, callback):
        """Callback for link buttons. Closes the popover then runs the actual action"""
        popover = button.get_parent().get_parent()
        popover.popdown()
        callback(button)

    def on_install_clicked(self, button):
        """Handler for installing service games"""
        self.service.install(self.db_game)

    def on_game_state_changed(self, game: Game) -> None:
        """Handler called when the game has changed state"""
        if (self.game.is_db_stored and game.id == self.game.id) or (self.appid and game.appid == self.appid):
            self.game = game
        elif self.game != game:
            return

        for child in self.get_children():
            child.destroy()
        self.update_view()



import os
import time
from gettext import gettext as _

from gi.repository import GObject, Gtk, Pango

from lutris.gui.dialogs import display_error
from lutris.util.downloader import Downloader
from lutris.util.jobs import schedule_repeating_at_idle
from lutris.util.log import logger
from lutris.util.strings import gtk_safe, human_size

# Same reason as Downloader
get_time = time.monotonic


class DownloadCollectionProgressBox(Gtk.Box):
    """Progress bar used to monitor a collection of files download."""

    max_retries = 3

    __gsignals__ = {
        "complete": (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_PYOBJECT,)),
        "cancel": (GObject.SignalFlags.RUN_LAST, None, ()),
        "error": (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_PYOBJECT,)),
    }

    def __init__(self, file_collection, cancelable=True, downloader=None):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)

        self.downloader = downloader
        self.is_complete = False
        self._file_queue = file_collection.files_list.copy()
        self._file_download = None  # file being downloaded
        self.title = file_collection.human_url
        self.num_files_downloaded = 0
        self.num_files_to_download = file_collection.num_files
        self.num_retries = 0
        self.full_size = file_collection.full_size
        self.current_size = 0
        self.time_left = "00:00:00"
        self.time_left_check_time = 0
        self.last_size = 0
        self.avg_speed = 0
        self.speed_list = []

        top_box = Gtk.Box()
        self.main_label = Gtk.Label(self.title)
        self.main_label.set_alignment(0, 0)
        self.main_label.set_property("wrap", True)
        self.main_label.set_margin_bottom(10)
        self.main_label.set_selectable(True)
        self.main_label.set_property("ellipsize", Pango.EllipsizeMode.MIDDLE)
        top_box.pack_start(self.main_label, False, False, 0)

        self.cancel_button = Gtk.Button.new_with_mnemonic(_("_Cancel"))
        self.cancel_cb_id = self.cancel_button.connect("clicked", self.on_cancel_clicked)
        if not cancelable:
            self.cancel_button.set_sensitive(False)
        top_box.pack_end(self.cancel_button, False, False, 0)

        self.pack_start(top_box, True, True, 0)

        self.file_name_label = Gtk.Label()
        self.file_name_label.set_alignment(0, 0)
        self.file_name_label.set_property("wrap", True)
        self.file_name_label.set_margin_bottom(10)
        self.file_name_label.set_selectable(True)
        self.file_name_label.set_property("ellipsize", Pango.EllipsizeMode.MIDDLE)
        self.pack_start(self.file_name_label, True, True, 0)

        progress_box = Gtk.Box()
        self.progressbar = Gtk.ProgressBar()
        self.progressbar.set_margin_top(5)
        self.progressbar.set_margin_bottom(5)
        self.progressbar.set_margin_right(10)
        progress_box.pack_start(self.progressbar, True, True, 0)

        self.pack_start(progress_box, False, False, 0)

        self.progress_label = Gtk.Label()
        self.progress_label.set_alignment(0, 0)
        self.pack_start(self.progress_label, True, True, 0)

        self.show_all()
        self.cancel_button.hide()

    def update_download_file_label(self, file_name):
        """Update file label to file being downloaded"""
        self.file_name_label.set_text(file_name)

    def get_next_file_from_queue(self):
        """Returns the file to download; if there isn't one this will pull the next one
        from the queue and return that. If there are none there, this returns None."""
        if not self._file_download:
            if not self._file_queue:
                return None

            self._file_download = self._file_queue.pop()
            self.num_retries = 0

        return self._file_download

    def start(self) -> None:
        """Start downloading a file."""
        file = self.get_next_file_from_queue()
        if not file:
            self.cancel_button.set_sensitive(False)
            self.is_complete = True
            self.emit("complete", {})
            return

        # Check if the file already exists in the cache and skip this download then
        if os.path.exists(file.dest_file):
            logger.info("File exists, skipping download: '%s'", file.dest_file)
            self.num_files_downloaded += 1
            self.current_size += os.path.getsize(file.dest_file)
            self._file_download = None
            self.start()
            return

        # Use a temporary file to avoid problems with partially downloaded files
        tmp_path = file.dest_file + ".tmp"
        file.tmp_file = tmp_path
        if os.path.exists(tmp_path):
            os.remove(tmp_path)

        self.update_download_file_label(file.filename)
        if not self.downloader:
            try:
                self.downloader = Downloader(file.url, file.tmp_file, referer=file.referer, overwrite=True)
            except RuntimeError as ex:
                display_error(ex, parent=self.get_toplevel())
                self.emit("cancel")
                return

        schedule_repeating_at_idle(self._progress, interval_seconds=0.5)
        self.cancel_button.show()
        self.cancel_button.set_sensitive(True)
        if not self.downloader.state == self.downloader.DOWNLOADING:
            self.downloader.start()

    def set_retry_button(self):
        """Transform the cancel button into a retry button"""
        self.cancel_button.set_label(_("Retry"))
        self.cancel_button.disconnect(self.cancel_cb_id)
        self.cancel_cb_id = self.cancel_button.connect("clicked", self.on_retry_clicked)
        self.cancel_button.set_sensitive(True)

    def on_retry_clicked(self, button):
        """Retry current download."""
        logger.debug("Retrying download")
        button.set_label(_("Cancel"))
        button.disconnect(self.cancel_cb_id)
        self.cancel_cb_id = button.connect("clicked", self.on_cancel_clicked)
        self.downloader.reset()
        self.start()

    def on_cancel_clicked(self, _widget=None):
        """Cancel the current download."""
        logger.debug("Download cancel requested")
        if self.downloader:
            self.downloader.cancel()
        self.cancel_button.set_sensitive(False)
        self.emit("cancel")

    def _progress(self) -> bool:
        """Show download progress of current file."""
        if self.downloader.state in [self.downloader.CANCELLED, self.downloader.ERROR]:
            self.progressbar.set_fraction(0)
            if self.downloader.state == self.downloader.CANCELLED:
                self._set_text(_("Download interrupted"))
                self.emit("cancel")
            else:
                if self.num_retries > self.max_retries:
                    self._set_text(str(self.downloader.error)[:80])
                    self.emit("error", self.downloader.error)
                    return False
                self.num_retries += 1
                if self.downloader:
                    self.downloader.reset()
                    self.start()
            return False
        downloaded_size = self.current_size + self.downloader.downloaded_size
        progress = 0
        if self.full_size > 0:
            progress = min(downloaded_size / self.full_size, 1)
        self.progressbar.set_fraction(progress)
        self.update_speed_and_time()
        megabytes = 1024 * 1024
        progress_text = _("{downloaded} / {size} ({speed:0.2f}MB/s), {time} remaining").format(
            downloaded=human_size(downloaded_size),
            size=human_size(self.full_size),
            speed=float(self.avg_speed) / megabytes,
            time=self.time_left,
        )
        self._set_text(progress_text)
        if self.downloader.state == self.downloader.COMPLETED:
            self.num_files_downloaded += 1
            self.current_size += self.downloader.downloaded_size
            os.rename(self._file_download.tmp_file, self._file_download.dest_file)
            # set file to None to get next one
            self._file_download = None
            self.downloader = None
            # start the downloader to a new file or finish
            self.start()
            return False
        return True

    def update_speed_and_time(self):
        """Update time left and average speed."""
        elapsed_time = get_time() - self.time_left_check_time
        if elapsed_time < 1:  # Minimum delay
            return

        if not self.downloader:
            self.time_left = "???"
            return

        downloaded_size = self.current_size + self.downloader.downloaded_size
        elapsed_size = downloaded_size - self.last_size
        self.last_size = downloaded_size

        speed = elapsed_size / elapsed_time
        # last 20 speeds
        if len(self.speed_list) >= 20:
            self.speed_list.pop(0)
        self.speed_list.append(speed)

        self.avg_speed = sum(self.speed_list) / len(self.speed_list)
        if self.avg_speed == 0:
            self.time_left = "???"
            return

        average_time_left = (self.full_size - downloaded_size) / self.avg_speed
        minutes, seconds = divmod(average_time_left, 60)
        hours, minutes = divmod(minutes, 60)
        self.time_left_check_time = get_time()
        self.time_left = "%d:%02d:%02d" % (hours, minutes, seconds)

    def _set_text(self, text):
        markup = "<span size='10000'>{}</span>".format(gtk_safe(text))
        self.progress_label.set_markup(markup)



import os
from collections import defaultdict
from gettext import gettext as _

from lutris import settings
from lutris.runners.runner import Runner
from lutris.util import system
from lutris.util.display import DISPLAY_MANAGER

AMIGAS = {
    "A500": {
        "name": _("Amiga 500"),
        "bios_sha1": [
            "891e9a547772fe0c6c19b610baf8bc4ea7fcb785",
            "c39bd9094d4e5f4e28c1411f3086950406062e87",
            "90933936cce43ca9bc6bf375662c076b27e3c458",
        ],
    },
    "A500+": {"name": _("Amiga 500+"), "bios_sha1": ["c5839f5cb98a7a8947065c3ed2f14f5f42e334a1"]},
    "A600": {"name": _("Amiga 600"), "bios_sha1": ["02843c4253bbd29aba535b0aa3bd9a85034ecde4"]},
    "A1200": {"name": _("Amiga 1200"), "bios_sha1": ["e21545723fe8374e91342617604f1b3d703094f1"]},
    "A3000": {"name": _("Amiga 3000"), "bios_sha1": ["f8e210d72b4c4853e0c9b85d223ba20e3d1b36ee"]},
    "A4000": {
        "name": _("Amiga 4000"),
        "bios_sha1": ["5fe04842d04a489720f0f4bb0e46948199406f49", "c3c481160866e60d085e436a24db3617ff60b5f9"],
    },
    "A1000": {"name": _("Amiga 1000"), "bios_sha1": ["11f9e62cf299f72184835b7b2a70a16333fc0d88"]},
    "CD32": {
        "name": _("Amiga CD32"),
        "bios_sha1": ["3525be8887f79b5929e017b42380a79edfee542d"],
        "bios_ext_sha1": ["5bef3d628ce59cc02a66e6e4ae0da48f60e78f7f"],
    },
    "CDTV": {
        "name": _("Commodore CDTV"),
        "bios_sha1": [
            "891e9a547772fe0c6c19b610baf8bc4ea7fcb785",
            "c39bd9094d4e5f4e28c1411f3086950406062e87",
            "90933936cce43ca9bc6bf375662c076b27e3c458",
        ],
        "bios_ext_sha1": ["7ba40ffa17e500ed9fed041f3424bd81d9c907be"],
    },
}


def get_bios_hashes():
    """Return mappings of sha1 hashes to Amiga models
    The first mapping contains the kickstarts and the second one, the extensions (for CD32/CDTV)
    """
    hashes = defaultdict(list)
    ext_hashes = defaultdict(list)
    for model, model_def in AMIGAS.items():
        for sha1_hash in model_def["bios_sha1"]:
            hashes[sha1_hash].append(model)
        if "bios_ext_sha1" in model_def:
            for sha1_hash in model_def["bios_ext_sha1"]:
                ext_hashes[sha1_hash].append(model)
    return hashes, ext_hashes


def scan_dir_for_bios(path):
    """Return a tuple of mappings of Amiga models and their corresponding kickstart file.

    Kickstart files must reside in `path`
    The first mapping contains the kickstarts and the second one, the extensions (for CD32/CDTV)
    """
    bios_sizes = [262144, 524288]
    hashes, ext_hashes = get_bios_hashes()
    found_bios = {}
    found_ext = {}
    incomplete_bios = []
    for file_name in os.listdir(path):
        abs_path = os.path.join(path, file_name)
        file_size = os.path.getsize(abs_path)
        if file_size not in bios_sizes:
            continue
        checksum = system.get_file_checksum(abs_path, "sha1")
        if checksum in hashes:
            for model in hashes[checksum]:
                found_bios[model] = abs_path
        if checksum in ext_hashes:
            for model in ext_hashes[checksum]:
                found_ext[model] = abs_path
    for model in found_bios:
        if "bios_ext_sha1" in AMIGAS[model] and model not in found_ext:
            incomplete_bios.append(model)
    found_bios = {k: v for k, v in found_bios.items() if k not in incomplete_bios}
    return found_bios, found_ext


class fsuae(Runner):
    human_name = _("FS-UAE")
    description = _("Amiga emulator")
    flatpak_id = "net.fsuae.FS-UAE"
    platforms = [
        AMIGAS["A500"]["name"],
        AMIGAS["A500+"]["name"],
        AMIGAS["A600"]["name"],
        AMIGAS["A1200"]["name"],
        AMIGAS["A3000"]["name"],
        AMIGAS["A4000"]["name"],
        AMIGAS["A1000"]["name"],
        AMIGAS["CD32"]["name"],
        AMIGAS["CDTV"]["name"],
    ]

    model_choices = [(model["name"], key) for key, model in AMIGAS.items()]

    cpumodel_choices = [
        (_("68000"), "68000"),
        (_("68010"), "68010"),
        (_("68020 with 24-bit addressing"), "68EC020"),
        (_("68020"), "68020"),
        (_("68030 without internal MMU"), "68EC030"),
        (_("68030"), "68030"),
        (_("68040 without internal FPU and MMU"), "68EC040"),
        (_("68040 without internal FPU"), "68LC040"),
        (_("68040 without internal MMU"), "68040-NOMMU"),
        (_("68040"), "68040"),
        (_("68060 without internal FPU and MMU"), "68EC060"),
        (_("68060 without internal FPU"), "68LC060"),
        (_("68060 without internal MMU"), "68060-NOMMU"),
        (_("68060"), "68060"),
        (_("Auto"), "auto"),
    ]
    memory_choices = [
        (_("0"), "0"),
        (_("1 MB"), "1024"),
        (_("2 MB"), "2048"),
        (_("4 MB"), "4096"),
        (_("8 MB"), "8192"),
    ]
    zorroiii_choices = [
        (_("0"), "0"),
        (_("1 MB"), "1024"),
        (_("2 MB"), "2048"),
        (_("4 MB"), "4096"),
        (_("8 MB"), "8192"),
        (_("16 MB"), "16384"),
        (_("32 MB"), "32768"),
        (_("64 MB"), "65536"),
        (_("128 MB"), "131072"),
        (_("256 MB"), "262144"),
        (_("384 MB"), "393216"),
        (_("512 MB"), "524288"),
        (_("768 MB"), "786432"),
        (_("1 GB"), "1048576"),
    ]
    flsound_choices = [
        ("0", "0"),
        ("25", "25"),
        ("50", "50"),
        ("75", "75"),
        ("100", "100"),
    ]
    gpucard_choices = [
        ("None", "None"),
        ("UAEGFX", "uaegfx"),
        ("UAEGFX Zorro II", "uaegfx-z2"),
        ("UAEGFX Zorro III", "uaegfx-z3"),
        ("Picasso II Zorro II", "picasso-ii"),
        ("Picasso II+ Zorro II", "picasso-ii+"),
        ("Picasso IV", "picasso-iv"),
        ("Picasso IV Zorro II", "picasso-iv-z2"),
        ("Picasso IV Zorro III", "picasso-iv-z3"),
    ]
    gpumem_choices = [
        (_("0"), "0"),
        (_("1 MB"), "1024"),
        (_("2 MB"), "2048"),
        (_("4 MB"), "4096"),
        (_("8 MB"), "8192"),
        (_("16 MB"), "16384"),
        (_("32 MB"), "32768"),
        (_("64 MB"), "65536"),
        (_("128 MB"), "131072"),
        (_("256 MB"), "262144"),
    ]
    flspeed_choices = [
        (_("Turbo"), "0"),
        ("100%", "100"),
        ("200%", "200"),
        ("400%", "400"),
        ("800%", "800"),
    ]
    runner_executable = "fs-uae/fs-uae"
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("Boot disk"),
            "default_path": "game_path",
            "help": _(
                "The main floppy disk file with the game data. \n"
                "FS-UAE supports floppy images in multiple file formats: "
                "ADF, IPF, DMS are the most common. ADZ (compressed ADF) "
                "and ADFs in zip files are a also supported.\n"
                "Files ending in .hdf will be mounted as hard drives and "
                "ISOs can be used for Amiga CD32 and CDTV models."
            ),
        },
        {
            "option": "disks",
            "section": _("Media"),
            "type": "multiple_file",
            "label": _("Additional floppies"),
            "default_path": "game_path",
            "help": _("The additional floppy disk image(s)."),
        },
        {
            "option": "cdrom_image",
            "section": _("Media"),
            "label": _("CD-ROM image"),
            "type": "file",
            "help": _("CD-ROM image to use on non CD32/CDTV models"),
        },
    ]

    runner_options = [
        {
            "option": "model",
            "label": _("Amiga model"),
            "type": "choice",
            "choices": model_choices,
            "default": "A500",
            "help": _("Specify the Amiga model you want to emulate."),
        },
        {
            "option": "kickstart_file",
            "section": _("Kickstart"),
            "label": _("Kickstart ROMs location"),
            "type": "file",
            "help": _(
                "Choose the folder containing original Amiga Kickstart "
                "ROMs. Refer to FS-UAE documentation to find how to "
                "acquire them. Without these, FS-UAE uses a bundled "
                "replacement ROM which is less compatible with Amiga "
                "software."
            ),
        },
        {
            "option": "kickstart_ext_file",
            "section": _("Kickstart"),
            "label": _("Extended Kickstart location"),
            "type": "file",
            "help": _("Location of extended Kickstart used for CD32"),
        },
        {
            "option": "gfx_fullscreen_amiga",
            "section": _("Graphics"),
            "label": _("Fullscreen (F12 + F to switch)"),
            "type": "bool",
            "default": False,
        },
        {
            "option": "scanlines",
            "section": _("Graphics"),
            "label": _("Scanlines display style"),
            "type": "bool",
            "default": False,
            "help": _("Activates a display filter adding scanlines to imitate the displays of yesteryear."),
        },
        {
            "option": "grafixcard",
            "section": _("Graphics"),
            "label": _("Graphics Card"),
            "type": "choice",
            "choices": gpucard_choices,
            "default": "None",
            "advanced": True,
            "help": _(
                "Use this option to enable a graphics card. This option is none by default, in "
                "which case only chipset graphics (OCS/ECS/AGA) support is available."
            ),
        },
        {
            "option": "grafixmemory",
            "section": _("Graphics"),
            "label": _("Graphics Card RAM"),
            "type": "choice",
            "choices": gpumem_choices,
            "default": "0",
            "advanced": True,
            "help": _(
                "Override the amount of graphics memory on the graphics card. The 0 MB option is "
                "not really valid, but exists for user interface reasons."
            ),
        },
        {
            "option": "cpumodel",
            "label": _("CPU"),
            "type": "choice",
            "choices": cpumodel_choices,
            "default": "auto",
            "advanced": True,
            "help": _(
                "Use this option to override the CPU model in the emulated Amiga. All Amiga "
                "models imply a default CPU model, so you only need to use this option if you "
                "want to use another CPU."
            ),
        },
        {
            "option": "fmemory",
            "label": _("Fast Memory"),
            "type": "choice",
            "choices": memory_choices,
            "default": "0",
            "advanced": True,
            "help": _("Specify how much Fast Memory the Amiga model should have."),
        },
        {
            "option": "ziiimem",
            "label": _("Zorro III RAM"),
            "type": "choice",
            "choices": zorroiii_choices,
            "default": "0",
            "advanced": True,
            "help": _(
                "Override the amount of Zorro III Fast memory, specified in KB. Must be a "
                "multiple of 1024. The default value depends on [amiga_model]. Requires a "
                "processor with 32-bit address bus, (use for example the A1200/020 model)."
            ),
        },
        {
            "option": "fdvolume",
            "section": _("Media"),
            "label": _("Floppy Drive Volume"),
            "type": "choice",
            "choices": flsound_choices,
            "default": "0",
            "advanced": True,
            "help": _("Set volume to 0 to disable floppy drive clicks when the drive is empty. Max volume is 100."),
        },
        {
            "option": "fdspeed",
            "section": _("Media"),
            "label": _("Floppy Drive Speed"),
            "type": "choice",
            "choices": flspeed_choices,
            "default": "100",
            "advanced": True,
            "help": _(
                "Set the speed of the emulated floppy drives, in percent. "
                "For example, you can specify 800 to get an 8x increase in "
                "speed. Use 0 to specify turbo mode. Turbo mode means that "
                "all floppy operations complete immediately. The default is 100 for most models."
            ),
        },
        {
            "option": "jitcompiler",
            "label": _("JIT Compiler"),
            "type": "bool",
            "default": False,
            "advanced": True,
        },
        {
            "option": "gamemode",
            "label": _("Feral GameMode"),
            "type": "bool",
            "default": False,
            "advanced": True,
            "help": _("Automatically uses Feral GameMode daemon if available. Set to true to disable the feature."),
        },
        {
            "option": "govwarning",
            "label": _("CPU governor warning"),
            "type": "bool",
            "default": False,
            "advanced": True,
            "help": _(
                "Warn if running with a CPU governor other than performance. Set to true to disable the warning."
            ),
        },
        {
            "option": "bsdsocket",
            "label": _("UAE bsdsocket.library"),
            "type": "bool",
            "default": False,
            "advanced": True,
        },
    ]

    @property
    def directory(self):
        return os.path.join(settings.RUNNER_DIR, "fs-uae")

    def get_platform(self):
        model = self.runner_config.get("model")
        if model:
            for index, machine in enumerate(self.model_choices):
                if machine[1] == model:
                    return self.platforms[index]
        return ""

    def get_absolute_path(self, path):
        """Return the absolute path for a file"""
        return path if os.path.isabs(path) else os.path.join(self.game_path, path)

    def insert_floppies(self):
        disks = []
        main_disk = self.game_config.get("main_file")
        if main_disk:
            disks.append(main_disk)

        game_disks = self.game_config.get("disks") or []
        for disk in game_disks:
            if disk not in disks:
                disks.append(disk)
        # Make all paths absolute
        disks = [self.get_absolute_path(disk) for disk in disks]
        drives = []
        floppy_images = []
        for drive, disk_path in enumerate(disks):
            disk_param = self.get_disk_param(disk_path)
            drives.append("--%s_%d=%s" % (disk_param, drive, disk_path))
            if disk_param == "floppy_drive":
                floppy_images.append("--floppy_image_%d=%s" % (drive, disk_path))
        cdrom_image = self.game_config.get("cdrom_image")
        if cdrom_image:
            drives.append("--cdrom_drive_0=%s" % self.get_absolute_path(cdrom_image))
        return drives + floppy_images

    def get_disk_param(self, disk_path):
        amiga_model = self.runner_config.get("model")
        if amiga_model in ("CD32", "CDTV"):
            return "cdrom_drive"
        if disk_path.lower().endswith(".hdf"):
            return "hard_drive"
        return "floppy_drive"

    def get_params(self):  # pylint: disable=too-many-branches
        params = []
        option_params = {
            "kickstart_file": "--kickstart_file=%s",
            "kickstart_ext_file": "--kickstart_ext_file=%s",
            "model": "--amiga_model=%s",
            "cpumodel": "--cpu=%s",
            "fmemory": "--fast_memory=%s",
            "ziiimem": "--zorro_iii_memory=%s",
            "fdvolume": "--floppy_drive_volume=%s",
            "fdspeed": "--floppy_drive_speed=%s",
            "grafixcard": "--graphics_card=%s",
            "grafixmemory": "--graphics_memory=%s",
        }
        for option, param in option_params.items():
            option_value = self.runner_config.get(option)
            if option_value:
                params.append(param % option_value)

        if self.runner_config.get("gfx_fullscreen_amiga"):
            width = DISPLAY_MANAGER.get_current_resolution()[0]
            params.append("--fullscreen")
            # params.append("--fullscreen_mode=fullscreen-window")
            params.append("--fullscreen_mode=fullscreen")
            params.append("--fullscreen_width=%s" % width)
        if self.runner_config.get("jitcompiler"):
            params.append("--jit_compiler=1")
        if self.runner_config.get("bsdsocket"):
            params.append("--bsdsocket_library=1")
        if self.runner_config.get("gamemode"):
            params.append("--game_mode=0")
        if self.runner_config.get("govwarning"):
            params.append("--governor_warning=0")
        if self.runner_config.get("scanlines"):
            params.append("--scanlines=1")
        return params

    def play(self):
        return {"command": self.get_command() + self.get_params() + self.insert_floppies()}



RECOMMENDED_CORES = {
    "mesen": {
        "platforms": ["nes"],
        "extensions": [
            ".nes",
            ".fds",
            ".unf",
            ".unif",
        ],
    },
    "gambatte": {
        "platforms": ["gameboy", "gameboy_color"],
        "bios": [
            {
                "filename": "gb_bios.bin",
                "description": "Game Boy BIOS - Optional",
                "md5sum": "32fbbd84168d3482956eb3c5051637f5",
                "required": False,
            },
            {
                "filename": "gbc_bios.bin",
                "description": "Game Boy Color BIOS - Optional",
                "md5sum": "dbfce9db9deaa2567f6a84fde55f9680",
                "required": False,
            },
        ],
        "extensions": [
            ".gb",
            ".gbc",
            ".dmg",
        ],
    },
    "snes9x": {
        "platforms": ["snes"],
        "extensions": [
            ".smc",
            ".sfc",
            ".swc",
            ".fig",
            ".bs",
            ".st",
        ],
    },
    "picodrive": {
        "platforms": ["master_system", "game_gear", "genesis", "mega_cd", "32x"],
        "extensions": [
            ".mdx",
            ".md",
            ".smd",
            ".gen",
            ".bin",
            ".cue",
            ".iso",
            ".sms",
            ".gg",
            ".sg",
            ".68k",
            ".chd",
        ],
    },
    "opera": {
        "platforms": ["3do"],
        "extensions": [
            ".iso",
            ".bin",
            ".chd",
            ".cue",
        ],
        "bios": [
            {"filename": "panafz1.bin", "description": "Panasonic FZ-1", "md5sum": "f47264dd47fe30f73ab3c010015c155b"},
            {
                "filename": "panafz10.bin",
                "description": "Panasonic FZ-10",
                "md5sum": "51f2f43ae2f3508a14d9f56597e2d3ce",
            },
            {
                "filename": "panafz10-norsa.bin",
                "description": "Panasonic FZ-10 [RSA Patch]",
                "md5sum": "1477bda80dc33731a65468c1f5bcbee9",
            },
            {
                "filename": "panafz10e-anvil.bin",
                "description": "Panasonic FZ-10-E [Anvil]",
                "md5sum": "a48e6746bd7edec0f40cff078f0bb19f",
            },
            {
                "filename": "panafz10e-anvil-norsa.bin",
                "description": "Panasonic FZ-10-E [Anvil RSA Patch]",
                "md5sum": "cf11bbb5a16d7af9875cca9de9a15e09",
            },
            {
                "filename": "panafz1j.bin",
                "description": "Panasonic FZ-1J",
                "md5sum": "a496cfdded3da562759be3561317b605",
            },
            {
                "filename": "panafz1j-norsa.bin",
                "description": "Panasonic FZ-1J [RSA Patch]",
                "md5sum": "f6c71de7470d16abe4f71b1444883dc8",
            },
            {
                "filename": "goldstar.bin",
                "description": "Goldstar GDO-101M",
                "md5sum": "8639fd5e549bd6238cfee79e3e749114",
            },
            {
                "filename": "sanyotry.bin",
                "description": "Sanyo IMP-21J TRY",
                "md5sum": "35fa1a1ebaaeea286dc5cd15487c13ea",
            },
            {
                "filename": "3do_arcade_saot.bin",
                "description": "Shootout At Old Tucson",
                "md5sum": "8970fc987ab89a7f64da9f8a8c4333ff",
            },
        ],
    },
    "mupen64plus_next": {"platforms": ["n64"], "extensions": [".z64", ".n64", ".v64"]},
}



# Standard Library
import os
from gettext import gettext as _

# Lutris Modules
from lutris import settings
from lutris.exceptions import MissingGameExecutableError
from lutris.runners.runner import Runner
from lutris.util import system


class mupen64plus(Runner):
    human_name = _("Mupen64Plus")
    description = _("Nintendo 64 emulator")
    platforms = [_("Nintendo 64")]
    runner_executable = "mupen64plus/mupen64plus"
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("ROM file"),
            "help": _("The game data, commonly called a ROM image."),
        }
    ]
    runner_options = [
        {
            "option": "fullscreen",
            "type": "bool",
            "label": _("Fullscreen"),
            "default": True,
        },
        {"option": "hideosd", "type": "bool", "label": _("Hide OSD"), "default": True},
    ]

    @property
    def working_dir(self):
        return os.path.join(settings.RUNNER_DIR, "mupen64plus")

    def play(self):
        arguments = self.get_command()
        if self.runner_config.get("hideosd"):
            arguments.append("--noosd")
        else:
            arguments.append("--osd")
        if self.runner_config.get("fullscreen"):
            arguments.append("--fullscreen")
        else:
            arguments.append("--windowed")
        rom = self.game_config.get("main_file") or ""
        if not system.path_exists(rom):
            raise MissingGameExecutableError(filename=rom)
        arguments.append(rom)
        return {"command": arguments}



import copy
import time
from typing import Any, Callable, Optional, Set

from lutris.database import games
from lutris.database.categories import (
    get_game_ids_for_categories,
    get_uncategorized_game_ids,
    normalized_category_names,
)
from lutris.exceptions import InvalidSearchTermError
from lutris.runners import get_runner_human_name
from lutris.search_predicate import (
    FLAG_TEXTS,
    TRUE_PREDICATE,
    AndPredicate,
    FlagPredicate,
    FunctionPredicate,
    MatchPredicate,
    NotPredicate,
    OrPredicate,
    SearchPredicate,
    TextPredicate,
)
from lutris.services import SERVICES
from lutris.util.strings import get_formatted_playtime, parse_playtime_parts
from lutris.util.tokenization import (
    TokenReader,
    clean_token,
    tokenize_search,
)

ISOLATED_TOKENS = set([":", "-", "(", ")", "<", ">", ">=", "<="])
ITEM_STOP_TOKENS = (ISOLATED_TOKENS | set(["OR", "AND"])) - set(["(", "-"])


def read_flag_token(tokens: TokenReader) -> Optional[bool]:
    token = tokens.get_cleaned_token() or ""
    folded = token.casefold()
    if folded in FLAG_TEXTS:
        return FLAG_TEXTS[folded]
    raise InvalidSearchTermError(f"'{token}' was found where a flag was expected.")


class BaseSearch:
    tags: Set[str] = set()

    def __init__(self, text: str) -> None:
        self.text = text
        self.predicate: Optional[SearchPredicate] = None

    def __str__(self) -> str:
        return self.text

    @property
    def is_empty(self) -> bool:
        return not self.text and not self.predicate

    def matches(self, candidate: Any) -> bool:
        return self.get_predicate().accept(candidate)

    def get_candidate_text(self, candidate: Any) -> str:
        return str(candidate)

    def has_component(self, component_name: str) -> bool:
        if component_name in self.tags:
            prev_token = None
            for token in tokenize_search(self.text, ISOLATED_TOKENS):
                if not token.isspace():
                    if token == ":" and prev_token and prev_token.casefold() == component_name:
                        return True
                    prev_token = token
        return False

    def get_predicate(self) -> SearchPredicate:
        if self.predicate is None:
            if self.text:
                raw_tokens = tokenize_search(self.text, ISOLATED_TOKENS)
                tokens = TokenReader(raw_tokens)
                self.predicate = self._parse_or(tokens) or TRUE_PREDICATE
            else:
                self.predicate = TRUE_PREDICATE
        return self.predicate

    def _parse_or(self, tokens: TokenReader) -> Optional[SearchPredicate]:
        parsed = self._parse_items(tokens)
        parts = []
        if parsed:
            parts.append(parsed)

            while tokens.consume("OR"):  # case-sensitive!
                more = self._parse_items(tokens)
                if not more:
                    break

                parts.append(more)

        if not parts:
            return None
        if len(parts) == 1:
            return parts[0]
        return OrPredicate(parts)

    def _parse_items(self, tokens: TokenReader) -> Optional[SearchPredicate]:
        buffer = []
        while True:
            parsed = self._parse_item(tokens)
            if parsed:
                buffer.append(parsed)
            else:
                break

        if not buffer:
            return None
        if len(buffer) == 1:
            return buffer[0]
        return AndPredicate(buffer)

    def _parse_item(self, tokens: TokenReader) -> Optional[SearchPredicate]:
        # AND is kinda fake - we and together items by default anyway,
        # so we'll just ignore this conjunction.
        while tokens.consume("AND"):
            pass

        token = tokens.peek_token()

        if not token or token in ITEM_STOP_TOKENS:
            return None

        if token.startswith('"'):
            tokens.get_token()  # consume token
            return self.get_text_predicate(clean_token(token))

        if tokens.consume("("):
            predicate = self._parse_or(tokens) or TRUE_PREDICATE
            tokens.consume(")")
            return predicate

        if tokens.consume("-"):
            inner = self._parse_items(tokens)
            if inner:
                return NotPredicate(inner)

        saved_index = tokens.index

        tokens.get_token()  # consume tag name
        if tokens.consume(":"):
            name = token.casefold()
            if name in self.tags:
                try:
                    return self.get_part_predicate(name, tokens)
                except InvalidSearchTermError:
                    pass

        # If the tag is no good, we'll rewind and fall back on a
        # literal text predicate for the whole thing
        tokens.index = saved_index

        text_token = tokens.get_cleaned_token_sequence(stop_function=self.is_stop_token)
        if text_token:
            return self.get_text_predicate(text_token)

        return None

    def with_predicate(self, predicate: SearchPredicate):
        old_predicate = self.get_predicate()  # force generation of predicate
        new_search = copy.copy(self)
        new_search.predicate = AndPredicate([old_predicate, predicate])
        return new_search

    def get_part_predicate(self, name: str, tokens: TokenReader) -> SearchPredicate:
        raise InvalidSearchTermError(f"'{name}' is not a valid search tag.")

    def get_text_predicate(self, text: str) -> SearchPredicate:
        return TextPredicate(text, self.get_candidate_text, tag="")

    def is_stop_token(self, tokens: TokenReader) -> bool:
        """This function decides when to stop when reading an item;
        pass this to tokens.get_cleaned_token_sequence().

        It will stop at the end of tokens, any of our stop tokens
        like AND, or at any tag, which must be a known tag followed by
        a colon."""
        peeked = tokens.peek_tokens(2)
        if not peeked:
            return True
        if peeked[0] in ITEM_STOP_TOKENS:
            return True
        if len(peeked) > 1 and peeked[1] == ":" and peeked[0].casefold() in self.tags:
            return True
        return False

    def quote_token(self, text: str) -> str:
        if text and " " not in text:
            tokens = list(tokenize_search(text, ISOLATED_TOKENS))
            test_reader = TokenReader(tokens)
            cleaned = test_reader.get_cleaned_token_sequence(self.is_stop_token)
            if cleaned == text:
                return text
        return f'"{text}"'


class GameSearch(BaseSearch):
    """A search for games, which applies to the games database dictionaries, not the Game objects."""

    tags = set(
        [
            "installed",
            "hidden",
            "favorite",
            "categorized",
            "category",
            "source",
            "service",  # an alias for source
            "runner",
            "platform",
            "playtime",
            "lastplayed",
            "directory",
        ]
    )

    def __init__(self, text: str, service=None) -> None:
        self.service = service
        super().__init__(text)

    def get_candidate_text(self, candidate: Any) -> str:
        return candidate["name"]

    def get_part_predicate(self, name: str, tokens: TokenReader) -> SearchPredicate:
        if name == "category":
            category = tokens.get_cleaned_token() or ""
            return self.get_category_predicate(category)

        if name in ("source", "service"):
            service_name = tokens.get_cleaned_token() or ""
            return self.get_service_predicate(service_name)

        if name == "runner":
            runner_name = tokens.get_cleaned_token() or ""
            return self.get_runner_predicate(runner_name)

        if name == "platform":
            platform = tokens.get_cleaned_token() or ""
            return self.get_platform_predicate(platform)

        if name == "playtime":
            return self.get_playtime_predicate(tokens)

        if name == "lastplayed":
            return self.get_lastplayed_predicate(tokens)

        if name == "directory":
            directory = tokens.get_cleaned_token_sequence(stop_function=self.is_stop_token) or ""
            return self.get_directory_predicate(directory)

        # All flags handle the 'maybe' option the same way, so we'll
        # group them at the end.
        flag = read_flag_token(tokens)

        flag_predicate = self.get_flag_predicate(name, flag)

        if flag_predicate:
            return flag_predicate

        return super().get_part_predicate(name, tokens)

    def get_flag_predicate(self, name: str, flag: Optional[bool]) -> Optional[SearchPredicate]:
        if name == "installed":
            return self.get_installed_predicate(flag)

        if name == "hidden":
            return self.get_category_flag_predicate(".hidden", "hidden", in_category=flag)

        if name == "favorite":
            return self.get_category_flag_predicate("favorite", "favorite", in_category=flag)

        if name == "categorized":
            return self.get_categorized_predicate(flag)

        return None

    def get_playtime_predicate(self, tokens: TokenReader) -> SearchPredicate:
        def get_game_playtime(db_game):
            return db_game.get("playtime")

        return self.get_duration_predicate(get_game_playtime, tokens, tag="playtime")

    def get_lastplayed_predicate(self, tokens: TokenReader) -> SearchPredicate:
        now = time.time()

        def get_game_lastplayed_duration_ago(db_game):
            lastplayed = db_game.get("lastplayed")
            if lastplayed:
                return (now - lastplayed) / (60 * 60)
            return None

        return self.get_duration_predicate(get_game_lastplayed_duration_ago, tokens, tag="lastplayed")

    def get_duration_predicate(self, value_function: Callable, tokens: TokenReader, tag: str) -> SearchPredicate:
        def match_greater_playtime(db_game):
            game_playtime = value_function(db_game)
            return game_playtime and game_playtime > duration

        def match_lesser_playtime(db_game):
            game_playtime = value_function(db_game)
            return game_playtime and game_playtime < duration

        def match_playtime(db_game):
            game_playtime = value_function(db_game)
            return game_playtime and duration_parts.matches(game_playtime)

        operator = tokens.peek_token()
        if operator == ">":
            matcher = match_greater_playtime
            tokens.get_token()
        elif operator == "<":
            matcher = match_lesser_playtime
            tokens.get_token()
        elif operator == ">=":
            matcher = lambda *a: match_greater_playtime(*a) or match_playtime(*a)  # noqa: E731
            tokens.get_token()
        elif operator == "<=":
            matcher = lambda *a: match_lesser_playtime(*a) or match_playtime(*a)  # noqa: E731
            tokens.get_token()
        else:
            matcher = match_playtime

        # We'll hope none of our tags are ever part of a legit duration
        duration_text = tokens.get_cleaned_token_sequence(stop_function=self.is_stop_token)
        if not duration_text:
            raise InvalidSearchTermError("A blank is not a valid duration.")

        try:
            duration_parts = parse_playtime_parts(duration_text)
            duration = duration_parts.get_total_hours()
        except ValueError as ex:
            raise InvalidSearchTermError(f"'{duration_text}' is not a valid playtime.") from ex

        text = f"{tag}:{operator}{get_formatted_playtime(duration)}"
        return FunctionPredicate(matcher, text)

    def get_directory_predicate(self, directory: str) -> SearchPredicate:
        return TextPredicate(directory, lambda c: c.get("directory"), tag="directory")

    def get_installed_predicate(self, installed: Optional[bool]) -> SearchPredicate:
        if self.service:

            def is_installed(db_game):
                appid = db_game.get("appid")
                return bool(appid and appid in games.get_service_games(self.service.id))

            return FlagPredicate(installed, is_installed, tag="installed")

        return FlagPredicate(installed, lambda db_game: bool(db_game["installed"]), tag="installed")

    def get_categorized_predicate(self, categorized: Optional[bool]) -> SearchPredicate:
        uncategorized_ids = get_uncategorized_game_ids()

        def is_categorized(db_game):
            return db_game["id"] not in uncategorized_ids

        return FlagPredicate(categorized, is_categorized, tag="categorized")

    def get_category_predicate(self, category: str) -> SearchPredicate:
        names = normalized_category_names(category, subname_allowed=True)
        category_game_ids = set(get_game_ids_for_categories(names))

        def match_category(db_game):
            game_id = db_game["id"]
            return game_id in category_game_ids

        text = f"category:{self.quote_token(category)}"
        return MatchPredicate(match_category, text=text, tag="category", value=category)

    def get_category_flag_predicate(self, category: str, tag: str, in_category: Optional[bool] = True) -> FlagPredicate:
        names = normalized_category_names(category, subname_allowed=True)
        category_game_ids = set(get_game_ids_for_categories(names))

        def is_in_category(db_game):
            game_id = db_game["id"]
            return game_id in category_game_ids

        return FlagPredicate(in_category, is_in_category, tag=tag)

    def get_service_predicate(self, service_name: str) -> SearchPredicate:
        service_name = service_name.casefold()
        text = f"source:{service_name}"

        if service_name == "none":

            def match_service(db_game):
                game_service = db_game.get("service")
                return not game_service
        else:

            def match_service(db_game):
                game_service = db_game.get("service")
                if not game_service:
                    return False

                if game_service.casefold() == service_name:
                    return True

                service = SERVICES.get(game_service)
                return service and service_name in service.name.casefold()

        return MatchPredicate(match_service, text=text, tag="source", value=service_name)

    def get_runner_predicate(self, runner_name: str) -> SearchPredicate:
        folded_runner_name = runner_name.casefold()
        text = f"runner:{self.quote_token(runner_name)}"

        if folded_runner_name == "none":

            def match_runner(db_game):
                game_runner = db_game.get("runner")
                return not game_runner
        else:

            def match_runner(db_game):
                game_runner = db_game.get("runner")

                if not game_runner:
                    return False

                if game_runner.casefold() == folded_runner_name:
                    return True

                runner_human_name = get_runner_human_name(game_runner)
                return runner_name in runner_human_name.casefold()

        return MatchPredicate(match_runner, text=text, tag="runner", value=runner_name)

    def get_platform_predicate(self, platform: str) -> SearchPredicate:
        folded_platform = platform.casefold()
        text = f"platform:{self.quote_token(platform)}"

        if folded_platform == "none":

            def match_platform(db_game):
                game_platform = db_game.get("platform")
                return not game_platform
        else:

            def match_platform(db_game):
                game_platform = db_game.get("platform")
                if game_platform:
                    return folded_platform in game_platform.casefold()
                if self.service:
                    platforms = [p.casefold() for p in self.service.get_game_platforms(db_game)]
                    matches = [p for p in platforms if folded_platform in p]
                    return any(matches)
                return False

        return MatchPredicate(match_platform, text=text, tag="platform", value=platform)


class RunnerSearch(BaseSearch):
    """A search for runners, which applies to the runner objects."""

    tags = set(["installed"])

    def get_candidate_text(self, candidate: Any) -> str:
        return f"{candidate.name}\n{candidate.description}"

    def get_part_predicate(self, name: str, tokens: TokenReader) -> SearchPredicate:
        if name == "installed":
            flag = read_flag_token(tokens)

            if flag is None:
                return TRUE_PREDICATE

            return self.get_installed_predicate(flag)

        return super().get_part_predicate(name, tokens)

    def get_installed_predicate(self, installed: bool) -> SearchPredicate:
        return FlagPredicate(installed, lambda runner: runner.is_installed(), tag="installed")



"""String utilities"""

import math
import re
import shlex
import time
import unicodedata
import uuid
from dataclasses import dataclass
from gettext import gettext as _
from typing import List, Optional, Tuple, Union

from gi.repository import GLib

from lutris.util.log import logger

NO_PLAYTIME = _("Never played")


def get_uuid_from_string(value: str) -> str:
    return str(uuid.uuid5(uuid.NAMESPACE_URL, str(value)))


def slugify(value: str) -> str:
    """Remove special characters from a string and slugify it.

    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.
    """
    _value = str(value)
    # This differs from the Lutris website implementation which uses the Django
    # version of `slugify` and uses the "NFKD" normalization method instead of
    # "NFD". This creates some inconsistencies in titles containing a trademark
    # symbols or some other special characters. The website version of slugify
    # will likely get updated to use the same normalization method.
    _value = unicodedata.normalize("NFD", _value).encode("ascii", "ignore")
    _value = _value.decode("utf-8")
    _value = str(re.sub(r"[^\w\s-]", "", _value)).strip().lower()
    slug = re.sub(r"[-\s]+", "-", _value)
    if not slug:
        # The slug is empty, likely because the string contains only non-latin
        # characters
        slug = get_uuid_from_string(value)
    return slug


def strip_accents(value: str) -> str:
    """Returns a string that is 'value', but with combining characters removed.

    This normalizes the text to form KD, which also removes any compatibility characters,
    so things like ellipsis are expanded to '...'.

    This also strips leading a trailing whitespace, and normalizes all remaining whitespace
    to single spaces.

    This does allow non-ASCII characters (like Greek or Cyrillic), and does not interfere with
    casing.

    We use this for a more forgiving search.
    """
    if value:
        decomposed = unicodedata.normalize("NFKD", value)
        result = ""
        prev_whitespace = False
        for ch in reversed(decomposed.strip()):
            combining = unicodedata.combining(ch)
            if not combining:
                whitespace = ch.isspace()
                if whitespace:
                    if not prev_whitespace:
                        result += " "
                else:
                    result += ch
                prev_whitespace = whitespace
        return result[::-1]  # We built the text backwards, so we must reverse it
    return value


def get_natural_sort_key(value: str, number_width: int = 16) -> str:
    """Returns a string with the numerical parts (runs of digits)
    0-padded out to 'number_width' digits."""

    def pad_numbers(text: str) -> str:
        return text.zfill(number_width) if text.isdigit() else text.casefold()

    runs = [pad_numbers(c) for c in re.split("([0-9]+)", value)]
    return "".join(runs)


def lookup_strings_in_text(string: str, text: str) -> List[str]:
    """Return each full line where a string was found in the multi-line text."""
    input_lines = text.split("\n")
    return [line for line in input_lines if string in line]


def parse_version(version: str) -> Tuple[List[int], str, str]:
    """Parse a version string

    Return a 3 element tuple containing:
     - The version number as a list of integers
     - The prefix (whatever characters before the version number)
     - The suffix (whatever comes after)

     Example::
        >>> parse_version("3.6-staging")
        ([3, 6], '', '-staging')

    Returns:
        tuple: (version number as list, prefix, suffix)
    """
    version_match = re.search(r"(\d[\d\.]+\d)", version)
    if not version_match:
        return [], "", ""
    version_number = version_match.groups()[0]
    prefix = version[0 : version_match.span()[0]]
    suffix = version[version_match.span()[1] :]
    return [int(p) for p in version_number.split(".")], suffix, prefix


def unpack_dependencies(string: str) -> List[Union[str, Tuple[str, ...]]]:
    """Parse a string to allow for complex dependencies
    Works in a similar fashion as Debian dependencies, separate dependencies
    are comma separated and multiple choices for satisfying a dependency are
    separated by pipes.

    Example: quake-steam | quake-gog, some-quake-mod returns:
        [('quake-steam', 'quake-gog'), 'some-quake-mod']
    """

    def _expand_dep(dep: str) -> Union[str, Tuple[str, ...]]:
        if "|" in dep:
            return tuple(option.strip() for option in dep.split("|") if option.strip())
        return dep.strip()

    if not string:
        return []
    return [dep for dep in [_expand_dep(dep) for dep in string.split(",")] if dep]


def gtk_safe(text: str) -> str:
    """Return a string ready to used in Gtk widgets, with anything that could
    be Pango markup escaped."""
    if not text:
        return ""

    return GLib.markup_escape_text(str(text))


def gtk_safe_urls(text: str) -> str:
    """Escapes the text as with gtk_safe, but detects URLs and converts them to
    anchor tags as well."""
    if not text:
        return ""

    parts = re.split(r"(http[s]?://(" r"?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+)", text)

    for index, part in enumerate(parts):
        if len(part) > 0:
            part = gtk_safe(part)
            if index % 2 != 0:  # every odd numbered part is the group in the regular expression
                part = f'<a href="{part}">{part}</a>'
            parts[index] = part

    return "".join(parts)


def is_valid_pango_markup(text: str) -> bool:
    def destroy_func(_user_data: None) -> None:
        pass  # required by GLib, but we don't need this callback

    if len(text) == 0:
        return True  # Trivial case - empty strings are always valid

    try:
        parser = GLib.MarkupParser()
        # DEFAULT_FLAGS == 0, but was not defined before GLib 2.74 so
        # we'll just hard-code the value.
        parser_flags: GLib.MarkupParseFlags = 0
        context = GLib.MarkupParseContext.new(
            parser=parser, flags=parser_flags, user_data=None, user_data_dnotify=destroy_func
        )

        markup = f"<markup>{text}</markup>"
        context.parse(markup, len(markup))
        return True
    except GLib.GError:  # type: ignore
        return False


def get_formatted_playtime(playtime: float) -> str:
    """Return a human-readable value of the play time"""
    if not playtime:
        return NO_PLAYTIME

    try:
        playtime = float(playtime)
    except ValueError:
        logger.warning("Invalid playtime value '%s'", playtime)
        return NO_PLAYTIME

    hours = math.floor(playtime)
    minutes = int(round((playtime - hours) * 60, 0))

    # If we're close enough to the next hour, we might wind up
    # with "x hours 60 minutes" and that looks dumb.
    if minutes >= 60:
        minutes -= 60
        hours += 1

    hours_unit = _("hour") if hours == 1 else _("hours")
    hours_text = f"{hours} {hours_unit}" if hours > 0 else ""
    minutes_unit = _("minute") if minutes == 1 else _("minutes")
    minutes_text = f"{minutes} {minutes_unit}" if minutes > 0 else ""

    formatted_time = " ".join([text for text in (hours_text, minutes_text) if text])
    if formatted_time:
        return formatted_time
    if playtime:
        return _("Less than a minute")
    return NO_PLAYTIME


def parse_playtime(text: str) -> float:
    """Parses a textual playtime into hours"""
    playtime = parse_playtime_parts(text)
    return playtime.get_total_hours()


@dataclass
class PlaytimeParts:
    years: float = 0.0
    months: float = 0.0
    weeks: float = 0.0
    days: float = 0.0
    hours: float = 0.0
    minutes: float = 0.0

    def is_empty(self) -> bool:
        return not (self.years or self.months or self.weeks or self.days or self.hours or self.minutes)

    def get_total_hours(self) -> float:
        return (
            self.hours
            + self.minutes / 60
            + self.days * 24
            + self.weeks * 24 * 7
            + self.months * 24 * 30
            + self.years * 24 * 365
        )

    def matches(self, hours: float) -> bool:
        total_hours = self.get_total_hours()
        if self.minutes != 0.0:
            return math.isclose(total_hours * 60, round(hours * 60, 0))
        if self.hours != 0.0:
            return math.isclose(total_hours, round(hours, 0))
        if self.days != 0.0:
            return math.isclose(total_hours / 24, round(hours / 24, 0))
        if self.weeks != 0.0:
            hours_per_week = 24 * 7
            return math.isclose(total_hours / hours_per_week, round(hours / hours_per_week, 0))
        if self.months != 0.0:
            hours_per_month = 24 * 30
            return math.isclose(total_hours / hours_per_month, round(hours / hours_per_month, 0))
        if self.years != 0.0:
            hours_per_year = 24 * 365
            return math.isclose(total_hours / hours_per_year, round(hours / hours_per_year, 0))

        # if 0 overall, treat as 0 minutes
        return math.isclose(total_hours * 60, round(hours * 60, 0))

    def add_part(self, num: float, unit: str) -> bool:
        # This function works out how many hours are meant by some
        # number of some unit.
        hour_units = ["h", "hr", "hours", "hour", _("hour"), _("hours")]
        minute_units = ["m", "min", "minute", "minutes", _("minute"), _("minutes")]
        day_units = ["d", _("day"), _("days")]
        week_units = ["wk", _("week"), _("weeks")]
        month_units = ["mo", _("month"), _("months")]
        year_units = ["yr", _("year"), _("years")]
        if unit in hour_units:
            self.hours += num
        elif unit in minute_units:
            self.minutes += num
        elif unit in day_units:
            self.days += num
        elif unit in week_units:
            self.weeks += num
        elif unit in month_units:
            self.months += num
        elif unit in year_units:
            self.years += num
        else:
            return False

        return True


def parse_playtime_parts(text: str) -> PlaytimeParts:
    text = text.strip().casefold()

    playtime = PlaytimeParts()

    if _("Less than a minute").casefold() == text:
        return playtime

    if NO_PLAYTIME.casefold() == text:
        return playtime

    error_message = _("'%s' is not a valid playtime.") % text

    # Handle a single number - assumed to be a count of hours
    try:
        playtime.hours = float(text)
        return playtime
    except ValueError:
        pass

    # Handle the easy case of "6:23".
    parts = text.split(":")
    if len(parts) == 2 and parts[0].strip().isdigit() and parts[1].strip().isdigit():
        try:
            playtime.hours = int(parts[0])
            playtime.minutes = int(parts[1])
            return playtime
        except ValueError as ex:
            raise ValueError(error_message) from ex

    # Handle the fancy format made of number unit pairts, like
    # "1 hour 23 minutes" or "2h57m"; we split this up into digit
    # and non-digit parts.
    parts = [p.strip() for p in re.split("([0-9.,]+)", text) if p and not p.isspace()]
    parts_iter = iter(parts)

    try:
        while True:
            num_text = next(parts_iter)
            try:
                num = float(num_text)
            except ValueError as ex:
                raise ValueError(error_message) from ex

            try:
                unit = next(parts_iter)
            except StopIteration as ex:
                if not playtime.is_empty():
                    unit = "minutes"
                else:
                    raise ValueError(error_message) from ex

            if not playtime.add_part(num, unit):
                raise ValueError(error_message)
    except StopIteration:
        pass

    return playtime


def _split_arguments(args: str, closing_quot: str = "", quotations: Optional[List[str]] = None) -> List[str]:
    if quotations is None:
        quotations = ["'", '"']
    try:
        return shlex.split(args + closing_quot)
    except ValueError as ex:
        message = ex.args[0]
        if message == "No closing quotation" and quotations:
            return _split_arguments(args, quotations[0], quotations[1:])
        logger.error(message)
        return []


def split_arguments(args: str) -> List[str]:
    """Wrapper around shlex.split that is more tolerant of errors"""
    if not args:
        # shlex.split seems to hangs when passed the None value
        return []
    return _split_arguments(args)


def human_size(size: float) -> str:
    """Shows a size in bytes in a more readable way"""
    units = ("bytes", "kB", "MB", "GB", "TB", "PB", "nuh uh", "no way", "BS")
    unit_index = 0
    while size > 1024:
        size = size / 1024
        unit_index += 1
    return "%0.1f %s" % (size, units[unit_index])


# inverse of human_size converts a human readable size to bytes
def computer_size(size: str) -> int:
    """Converts a human readable size to bytes"""
    units = ("bytes", "kb", "mb", "gb", "tb", "pb")
    unit_index = 0
    size = size.strip()
    size = size.casefold()
    for unit in units:
        if size.endswith(unit):
            size = size[: -len(unit)].strip()
            break
        unit_index += 1
    try:
        return int(float(size) * (1024**unit_index))
    except ValueError:
        logger.warning("Invalid size value '%s'", size)
        return 0  # if we can't parse the size, return 0 bytes to avoid crashing


def time_ago(timestamp: float) -> str:
    time_delta = time.time() - timestamp

    original_time_delta = time_delta
    if time_delta < 0:
        return _("in the future")
    if time_delta < 5:
        return _("just now")
    parts = []
    day_in_seconds = 3600 * 24
    hour_in_seconds = 3600
    days = 0
    hours = 0
    if time_delta >= 2 * day_in_seconds:
        days = int(time_delta // day_in_seconds)
        time_delta = time_delta - days * day_in_seconds
        parts.append(_("%d days") % days)
    if time_delta > 2 * hour_in_seconds:
        hours = int(time_delta // hour_in_seconds)
        time_delta = time_delta - hours * hour_in_seconds
        parts.append(_("%d hours") % hours)
    if not days and hours < 5 and time_delta > 60:
        minutes = int(time_delta // 60)
        time_delta = time_delta - minutes * 60
        if minutes != 1:
            parts.append(_("%d minutes") % minutes)
        else:
            parts.append(_("1 minute"))
    if original_time_delta < 90:
        seconds = int(time_delta)
        if seconds != 1:
            parts.append(_("%d seconds") % seconds)
        else:
            parts.append(_("1 second"))

    parts.append(_("ago"))
    return " ".join(parts)



from gettext import gettext as _

from gi.repository import GObject

from lutris.config import LutrisConfig
from lutris.gui.config.game_common import GameDialogCommon
from lutris.runners import get_runner_human_name


class RunnerConfigDialog(GameDialogCommon):
    """Runner config edit dialog."""

    __gsignals__ = {
        "runner-updated": (GObject.SIGNAL_RUN_FIRST, None, (str,)),
    }

    def __init__(self, runner, parent=None):
        super().__init__(_("Configure %s") % runner.human_name, config_level="runner", parent=parent)
        self.runner_name = runner.__class__.__name__
        self.saved = False
        self.lutris_config = LutrisConfig(runner_slug=self.runner_name)
        self.build_notebook()
        self.build_tabs()
        self.show_all()

    def get_search_entry_placeholder(self):
        return _("Search %s options") % get_runner_human_name(self.runner_name)

    def on_save(self, _wigdet, data=None):
        self.lutris_config.save()
        self.emit("runner-updated", self.runner_name)
        self.destroy()



# pylint: disable=no-member
# pylint:disable=using-constant-test
# pylint:disable=comparison-with-callable
from gettext import gettext as _
from math import floor

import gi

from lutris.util.jobs import schedule_at_idle
from lutris.util.log import logger

gi.require_version("PangoCairo", "1.0")

import cairo
from gi.repository import Gdk, GObject, Gtk, Pango, PangoCairo

from lutris.gui.widgets.utils import (
    MEDIA_CACHE_INVALIDATED,
    get_default_icon_path,
    get_runtime_icon_path,
    get_scaled_surface_by_path,
    get_surface_size,
)
from lutris.services.service_media import resolve_media_path
from lutris.util.path_cache import MISSING_GAMES

_MEDIA_CACHE_GENERATION_NUMBER = 0


class GridViewCellRendererText(Gtk.CellRendererText):
    """CellRendererText adjusted for grid view display, removes extra padding
    and caches cell metrics for improved resize performance."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.props.alignment = Pango.Alignment.CENTER
        self.props.wrap_mode = Pango.WrapMode.WORD
        self.props.xalign = 0.5
        self.props.yalign = 0
        self.fixed_width = None
        self.cached_height = {}
        self.cached_width = {}

    def set_width(self, width):
        self.fixed_width = width
        self.props.wrap_width = width
        self.clear_caches()

    def clear_caches(self):
        self.cached_height.clear()
        self.cached_width.clear()

    def do_get_preferred_width(self, widget):
        text = self.props.text  # pylint:disable=no-member
        if self.fixed_width and text in self.cached_width:
            return self.cached_width[text]

        width = Gtk.CellRendererText.do_get_preferred_width(self, widget)

        if self.fixed_width:
            self.cached_width[text] = width

        return width

    def do_get_preferred_width_for_height(self, widget, width):
        text = self.props.text  # pylint:disable=no-member
        if self.fixed_width and text in self.cached_width:
            return self.cached_width[text]

        width = Gtk.CellRendererText.do_get_preferred_width_for_height(self, widget, width)

        if self.fixed_width:
            self.cached_width[text] = width

        return width

    def do_get_preferred_height(self, widget):
        text = self.props.text  # pylint:disable=no-member
        if self.fixed_width and text in self.cached_height:
            return self.cached_height[text]

        height = Gtk.CellRendererText.do_get_preferred_height(self, widget)

        if self.fixed_width:
            self.cached_height[text] = height

        return height

    def do_get_preferred_height_for_width(self, widget, width):
        text = self.props.text  # pylint:disable=no-member
        if self.fixed_width and text in self.cached_height:
            return self.cached_height[text]

        height = Gtk.CellRendererText.do_get_preferred_height_for_width(self, widget, width)

        if self.fixed_width:
            self.cached_height[text] = height

        return height


class GridViewCellRendererImage(Gtk.CellRenderer):
    """A pixbuf cell renderer that takes not the pixbuf but a path to an image file;
    it loads that image only when rendering. It also has properties for its width
    and height, so it need not load the pixbuf to know its size."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._game_id = None
        self._service = None
        self._media_paths = []
        self._show_badges = True
        self._platform = None
        self._is_installed = True
        self.cached_surfaces_new = {}
        self.cached_surfaces_old = {}
        self.cached_surfaces_loaded = 0
        self.cached_surface_generation = 0
        self.badge_size = 0, 0
        self.badge_alpha = 0.6
        self.badge_fore_color = 1, 1, 1
        self.badge_back_color = 0, 0, 0
        self._inset_fractions = {}
        self._expected_width = None
        self._expected_height = None

    def inset_game(self, game_id: str, fraction: float) -> bool:
        """This function indicates that a particular game should be displayed inset by a certain fraction of
        its total size; 0 is full size, 0.1 would show it at 90% size, but centered.

        This is not bound as an attribute; it's used for an ephemeral animation, and we wouldn't want
        to mess with the GameStore to do it. Instead, the cell renderer tracks these per game ID, and
        the caller uses queue_draw() to trigger a redraw.

        Set the fraction to 0 for a game to remove the effect when done.

        This returns True if it alters the inset of a game, and False if not because it was
        already set that way."""
        if fraction > 0.0:
            if fraction != self._inset_fractions.get(game_id):
                self._inset_fractions[game_id] = fraction
                return True
        elif game_id in self._inset_fractions:
            del self._inset_fractions[game_id]
            return True

        return False

    def is_library_view(self):
        """Returns True if this is rendering for the library view (not a service view)"""
        return self.service is None

    def set_expected_size(self, width, height):
        """Sets the expected dimensions for library view based on the selected zoom level"""
        self._expected_width = width
        self._expected_height = height

    @GObject.Property(type=str)
    def game_id(self):
        """This is the path to the media file to be displayed."""
        return self._game_id

    @game_id.setter
    def game_id(self, value):
        self._game_id = value

    @GObject.Property(type=GObject.TYPE_PYOBJECT)
    def service(self):
        return self._service

    @service.setter
    def service(self, value):
        self._service = value

    @GObject.Property(type=GObject.TYPE_PYOBJECT)
    def media_paths(self):
        """This is the list of paths where the media to be displayed may be."""
        return self._media_paths

    @media_paths.setter
    def media_paths(self, value):
        self._media_paths = value

    @GObject.Property(type=bool, default=True)
    def show_badges(self):
        """This is the path to the media file to be displayed."""
        return self._show_badges

    @show_badges.setter
    def show_badges(self, value):
        self._show_badges = value

    @GObject.Property(type=str)
    def platform(self):
        """This is the platform text, a comma separated list; we try to convert
        this into icons, if it is not None."""
        return self._platform

    @platform.setter
    def platform(self, value):
        self._platform = value

    @GObject.Property(type=bool, default=True)
    def is_installed(self):
        """This flag indicates if the game is installed; if not the media is shown
        faded out."""
        return self._is_installed

    @is_installed.setter
    def is_installed(self, value):
        self._is_installed = value

    def _get_preferred_size(self):
        paths = self.media_paths
        if paths:
            path = paths[0]
            return path.width, path.height
        return 0, 0

    def do_get_preferred_width(self, widget):
        if self.is_library_view() and self._expected_width:
            return self._expected_width, self._expected_width
        size = self._get_preferred_size()
        return size[0], size[0]

    def do_get_preferred_height(self, widget):
        if self.is_library_view() and self._expected_height:
            return self._expected_height, self._expected_height
        size = self._get_preferred_size()
        return size[1], size[1]

    def do_render(self, cr, widget, background_area, cell_area, flags):
        media_path = resolve_media_path(self.media_paths) if self.media_paths else None
        if not media_path:
            return

        media_width = media_path.width
        media_height = media_path.height
        path = media_path.path
        alpha = 1 if self.is_installed else 100 / 255

        if media_width > 0 and media_height > 0 and path:
            surface = self._get_cached_surface_by_path(widget, path, size=(media_width, media_height))
            if not surface:
                # The default icon needs to be scaled to fill the cell space.
                path = get_default_icon_path((media_width, media_height))
                surface = self._get_cached_surface_by_path(
                    widget, path, size=(media_width, media_height), preserve_aspect_ratio=False
                )
            if surface:
                media_area = self.get_media_area(surface, cell_area)
                self.select_badge_metrics(surface, media_width, media_height)

                cr.save()

                # Adjust context to place media_area at 0,0 and scale it.
                inset_fraction = self._inset_fractions.get(self.game_id) or 0.0 if self.game_id else 0.0
                if inset_fraction > 0:
                    media_area.x += (media_area.width * inset_fraction) / 2
                    media_area.y += (media_area.height * inset_fraction) / 2
                    cr.translate(media_area.x, media_area.y)
                    cr.scale(1 - inset_fraction, 1 - inset_fraction)
                else:
                    cr.translate(media_area.x, media_area.y)

                media_area.x = 0
                media_area.y = 0

                if alpha >= 1:
                    self.render_media(cr, widget, surface, 0, 0)
                    if self.show_badges:
                        self._render_badges(cr, widget, surface, media_area)
                else:
                    cr.push_group()
                    self.render_media(cr, widget, surface, 0, 0)
                    if self.show_badges:
                        self._render_badges(cr, widget, surface, media_area)
                    cr.pop_group_to_source()
                    cr.paint_with_alpha(alpha)
                cr.restore()

                if inset_fraction > 0:
                    # Since we are restoring a scale change _again_ we need
                    # to update the layout with that change too, or the other text
                    # in the view will be broken. So lame.
                    layout = widget.create_pango_layout("")
                    PangoCairo.update_layout(cr, layout)

            # Idle time will wait until the widget has drawn whatever it wants to;
            # we can then discard surfaces we aren't using anymore.
            schedule_at_idle(self.cycle_cache)

    def select_badge_metrics(self, surface, media_width, media_height):
        """Updates fields holding data about the appearance of the badges;
        this sets self.badge_size to None if no badges should be shown at all."""

        def get_badge_icon_size():
            """Returns the size of the badge icons to render, or None to hide them. We check
            width for the smallest size because Dolphin has very thin banners, but we only hide
            badges for icons, not banners."""
            if media_width < 64:
                return None
            if media_height < 128:
                return 16, 16
            if media_height < 256:
                return 24, 24
            return 32, 32

        self.badge_size = get_badge_icon_size()
        on_bright_surface = self.badge_size and GridViewCellRendererImage.is_bright_corner(surface, self.badge_size)

        bright_color = 0.8, 0.8, 0.8
        dark_color = 0.2, 0.2, 0.2
        self.badge_fore_color = dark_color if on_bright_surface else bright_color
        self.badge_back_color = bright_color if on_bright_surface else dark_color

    @staticmethod
    def is_bright_corner(surface, corner_size):
        """Tests several pixels near the corner of the surface where the badges
        are drawn. If all are 'bright', we'll render the badges differently. This
        means all 4 components must be at least 128/255."""
        surface_format = surface.get_format()

        # We only use the ARGB32 format, so we just give up
        # for anything else.
        if surface_format != cairo.FORMAT_ARGB32:  # pylint:disable=no-member
            return False

        # Scale the corner according to the surface's scale factor -
        # normally the same as our UI scale factor.
        device_scale_x, device_scale_y = surface.get_device_scale()
        corner_pixel_width = int(corner_size[0] * device_scale_x)
        corner_pixel_height = int(corner_size[1] * device_scale_y)
        pixel_width = surface.get_width()
        pixel_height = surface.get_height()

        def is_bright_pixel(x, y):
            # Checks if a pixel is 'bright'; this does not care
            # if the pixel is big or little endian- it just checks
            # all four channels.
            if 0 <= x < pixel_width and 0 <= y < pixel_height:
                stride = surface.get_stride()
                data = surface.get_data()

                offset = (y * stride) + x * 4
                pixel = data[offset : offset + 4]

                for channel in pixel:
                    if channel < 128:
                        return False
                return True
            return False

        return (
            is_bright_pixel(pixel_width - 1, pixel_height - 1)
            and is_bright_pixel(pixel_width - corner_pixel_width, pixel_height - 1)
            and is_bright_pixel(pixel_width - 1, pixel_height - corner_pixel_height)
            and is_bright_pixel(pixel_width - corner_pixel_width, pixel_height - corner_pixel_height)
        )

    def get_media_area(self, surface, cell_area):
        """Computes the position of the upper left corner where we will render
        a surface within the cell area."""
        media_area = Gdk.Rectangle()
        width, height = get_surface_size(surface)

        # Horizontal centering (same for both library and service views)
        media_area.x = round(cell_area.x + (cell_area.width - width) / 2)

        if self.is_library_view():
            # For library view, center vertically within cell height
            media_area.y = round(cell_area.y + (cell_area.height - height) / 2)
        else:
            # Service views: bottom-aligned
            media_area.y = round(cell_area.y + cell_area.height - height)

        media_area.width, media_area.height = width, height
        return media_area

    def render_media(self, cr, widget, surface, x, y):
        """Renders the media itself, given the surface containing it
        and the position."""
        width, height = get_surface_size(surface)

        cr.set_source_surface(surface, x, y)
        cr.get_source().set_extend(cairo.Extend.PAD)  # pylint: disable=no-member
        cr.rectangle(x, y, width, height)
        cr.fill()

    def _render_badges(self, cr, widget, surface, media_area):
        self.render_platforms(cr, widget, surface, 0, media_area)

        game_id = self.game_id
        if game_id:
            if self.service:
                game_id = self.service.resolve_game_id(game_id)

            if game_id in MISSING_GAMES.missing_game_ids:
                self.render_text_badge(cr, widget, _("Missing"), 0, media_area.y + media_area.height)

    def render_platforms(self, cr, widget, surface, surface_x, media_area):
        """Renders the stack of platform icons."""
        platform = self.platform
        if platform and self.badge_size:
            icon_paths = self.get_platform_icon_paths(platform)
            if icon_paths:
                self.render_badge_stack(cr, widget, surface, surface_x, icon_paths, media_area)

    @staticmethod
    def get_platform_icon_paths(platform):
        if platform in GridViewCellRendererImage._platform_icon_paths:
            return GridViewCellRendererImage._platform_icon_paths[platform]

        if "," in platform:
            platforms = platform.split(",")  # pylint:disable=no-member
        else:
            platforms = [platform]

        icon_paths = []
        for p in platforms:
            icon_path = get_runtime_icon_path(p + "-symbolic")
            if icon_path:
                icon_paths.append(icon_path)

        GridViewCellRendererImage._platform_icon_paths[platform] = icon_paths
        return icon_paths

    _platform_icon_paths = {}

    def render_badge_stack(self, cr, widget, surface, surface_x, icon_paths, media_area):
        """Renders a vertical stack of badges, placed at the edge of the media, just to the left
        of 'media_area.right'. The icons in icon_paths are drawn from top to bottom, and spaced
        to fit in 'media_area', even if they overlap because of this."""

        badge_width = self.badge_size[0]
        badge_height = self.badge_size[1]
        alpha = self.badge_alpha
        fore_color = self.badge_fore_color
        back_color = self.badge_back_color

        def render_badge(badge_x, badge_y, path):
            cr.rectangle(badge_x, badge_y, badge_width, badge_height)
            cr.set_source_rgba(back_color[0], back_color[1], back_color[2], alpha)
            cr.fill()

            icon = self._get_cached_surface_by_path(widget, path, size=self.badge_size)
            cr.set_source_rgba(fore_color[0], fore_color[1], fore_color[2], alpha)
            cr.mask_surface(icon, badge_x, badge_y)

        media_right = surface_x + get_surface_size(surface)[0]

        x = media_right - badge_width
        spacing = (media_area.height - badge_height * len(icon_paths)) / max(1, len(icon_paths) - 1)
        spacing = min(spacing, 1)
        y_offset = floor(badge_height + spacing)
        y = media_area.y + media_area.height - badge_height - y_offset * (len(icon_paths) - 1)

        for icon_path in icon_paths:
            render_badge(x, y, icon_path)
            y = y + y_offset

    def render_text_badge(self, cr, widget, text, left, bottom):
        """Draws a short text in the lower left corner of the media, in the
        style of a badge."""

        def get_layout():
            """Constructs a layout with the text to draw, but also returns its size
            in pixels. This is boldfaced, but otherwise in the default font."""
            lo = widget.create_pango_layout(text)
            font = lo.get_context().get_font_description()
            font.set_weight(Pango.Weight.BOLD)
            lo.set_font_description(font)
            _, text_bounds = lo.get_extents()
            return lo, text_bounds.width / Pango.SCALE, text_bounds.height / Pango.SCALE

        if self.badge_size:
            alpha = self.badge_alpha
            fore_color = self.badge_fore_color
            back_color = self.badge_back_color

            layout, text_width, text_height = get_layout()

            cr.save()

            # To get the text to be as tall as a badge, we'll scale it
            # with Cairo. Scaling the font size does not work; the font
            # size measures the wrong height for this.
            text_scale = self.badge_size[1] / text_height
            text_height = self.badge_size[1]
            text_width = round(text_width * text_scale)

            cr.rectangle(left, bottom - text_height, text_width + 4, text_height)
            cr.set_source_rgba(back_color[0], back_color[1], back_color[2], alpha)
            cr.fill()

            cr.translate(left + 2, bottom - text_height)
            cr.scale(text_scale, text_scale)
            cr.set_source_rgba(fore_color[0], fore_color[1], fore_color[2], alpha)
            PangoCairo.update_layout(cr, layout)
            PangoCairo.show_layout(cr, layout)

            cr.restore()

            # Looks like we need to make cr.restore() take effect for
            # explicitly, or further text in this cairo context winds up scaled.
            # It must be doing something squirrely with the context that we just
            # spoiled with cr.restore(), and this fixes that.
            PangoCairo.update_layout(cr, layout)

    def clear_cache(self):
        """Discards all cached surfaces; used when some properties are changed."""
        self.cached_surfaces_old.clear()
        self.cached_surfaces_new.clear()

    def cycle_cache(self) -> None:
        """Is the key cache size control trick. When called, the surfaces cached or used
        since the last call are preserved, but those not touched are discarded.

        We call this at idle time after rendering a cell; this should keep all the surfaces
        rendered at that time, so during scrolling the visible media are kept and scrolling is smooth.
        At other times we may discard almost all surfaces, saving memory.

        We skip clearing anything if no surfaces have been loaded; this happens if drawing was
        serviced entirely from cache. GTK may have redrawn just one image or something, so
        let's not disturb the cache for that."""
        if self.cached_surfaces_loaded > 0:
            self.cached_surfaces_old = self.cached_surfaces_new
            self.cached_surfaces_new = {}
            self.cached_surfaces_loaded = 0

    def _get_cached_surface_by_path(self, widget, path, size, preserve_aspect_ratio=True):
        """This obtains the scaled surface to rander for a given media path; this is cached
        in this render, but we'll clear that cache when the media generation number is changed,
        or certain properties are. We also age surfaces from the cache at idle time after
        rendering."""
        if self.cached_surface_generation != _MEDIA_CACHE_GENERATION_NUMBER:
            self.cached_surface_generation = _MEDIA_CACHE_GENERATION_NUMBER
            self.clear_cache()

        key = widget, path, size, preserve_aspect_ratio

        if key in self.cached_surfaces_new:
            return self.cached_surfaces_new[key]

        if key in self.cached_surfaces_old:
            surface = self.cached_surfaces_old[key]
        else:
            surface = self._get_surface_by_path(widget, path, size, preserve_aspect_ratio)
            if surface:
                # We cache missing surfaces too, but only a successful load trigger
                # cache cycling
                self.cached_surfaces_loaded += 1

        self.cached_surfaces_new[key] = surface
        return surface

    def _get_surface_by_path(self, widget, path, size, preserve_aspect_ratio=True):
        cell_size = size
        scale_factor = widget.get_scale_factor() if widget else 1
        try:
            return get_scaled_surface_by_path(
                path, cell_size, scale_factor, preserve_aspect_ratio=preserve_aspect_ratio
            )
        except Exception as ex:
            # We need to survive nasty data in the media files, so the user can replace
            # them.
            logger.exception("Unable to load media '%s': %s", path, ex)
            return None


def _on_media_cached_invalidated() -> None:
    # Increment a counter, so we can passively detect when the media cache is invalid
    # without a per-object handler. We have no way to unregister that handler, but we never
    # need to unregister this global one.
    global _MEDIA_CACHE_GENERATION_NUMBER
    _MEDIA_CACHE_GENERATION_NUMBER += 1


MEDIA_CACHE_INVALIDATED.register(_on_media_cached_invalidated)



from gettext import gettext as _

from gi.repository import GObject, Gtk

from lutris.exceptions import InvalidGameMoveError
from lutris.game import GAME_UPDATED
from lutris.gui.dialogs import ModelessDialog, WarningDialog, display_error
from lutris.util.jobs import AsyncCall, schedule_repeating_at_idle
from lutris.util.path_cache import remove_from_path_cache
from lutris.util.strings import gtk_safe


class MoveDialog(ModelessDialog):
    __gsignals__ = {
        "game-moved": (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    def __init__(self, game, destination: str, parent: Gtk.Window = None) -> None:
        super().__init__(parent=parent, border_width=24)

        self.game = game
        self.destination = destination
        self.new_directory = None

        self.set_size_request(320, 60)
        self.set_decorated(False)
        vbox = Gtk.Box.new(Gtk.Orientation.VERTICAL, 12)
        label = Gtk.Label(label=_("Moving %s to %s..." % (game, destination)))
        vbox.add(label)
        self.progress = Gtk.ProgressBar(visible=True)
        self.progress.set_pulse_step(0.1)
        vbox.add(self.progress)
        self.get_content_area().add(vbox)
        self.progress_source_task = schedule_repeating_at_idle(self.show_progress, interval_seconds=0.125)
        self.connect("destroy", self.on_destroy)
        self.show_all()

    def on_destroy(self, _dialog):
        self.progress_source_task.unschedule()

    def move(self):
        AsyncCall(self._move_game, self._move_game_cb)

    def show_progress(self) -> bool:
        self.progress.pulse()
        return True

    def _move_game(self):
        # remove the old path from the cache
        remove_from_path_cache(self.game)
        # not safe to fire a signal from a thread- it will surely hit GTK and may crash
        self.new_directory = self.game.move(self.destination, no_signal=True)

    def _move_game_cb(self, _result, error):
        if error and isinstance(error, InvalidGameMoveError):
            secondary = _(
                "Do you want to change the game location anyway? No files can be moved, "
                "and the game configuration may need to be adjusted."
            )
            dlg = WarningDialog(message_markup=gtk_safe(error), secondary=secondary, parent=self)
            if dlg.result == Gtk.ResponseType.OK:
                self.new_directory = self.game.set_location(self.destination)
                self.on_game_moved(None, None)
            else:
                self.destroy()
            return

        self.on_game_moved(_result, error)

    def on_game_moved(self, _result, error):
        if error:
            display_error(error, parent=self)
        GAME_UPDATED.fire(self.game)  # because we could not fire this on the thread
        self.emit("game-moved")
        self.destroy()



"""Process management"""

import os
import shlex
import sys

from lutris.util.process import Process

# Processes that are considered sufficiently self-managing by the
# monitoring system. These are not considered game processes for
# the purpose of determining if a game is still running and Lutris
# will never attempt to send signals to these processes.
# This is mostly a minor UX improvement where wine games will exit
# faster if we let the wine processes tear themselves down.
SYSTEM_PROCESSES = {
    "wineserver",
    "services.exe",
    "winedevice.exe",
    "plugplay.exe",
    "explorer.exe",
    "wineconsole",
    "svchost.exe",
    "rpcss.exe",
    "rundll32.exe",
    "mscorsvw.exe",
    "iexplore.exe",
    "winedbg.exe",
    "tabtip.exe",
    "conhost.exe",
}


class ProcessWatcher:
    """Keeps track of child processes of the client"""

    def __init__(self, include_processes, exclude_processes):
        """Create a process watcher.
        Params:
            exclude_processes (str or list): list of processes that shouldn't be monitored
            include_processes (str or list): list of process that should be forced to be monitored
        """
        self.unmonitored_processes = (
            self.parse_process_list(exclude_processes) | SYSTEM_PROCESSES
        ) - self.parse_process_list(include_processes)

    @staticmethod
    def parse_process_list(process_list):
        """Parse a process list that may be given as a string"""
        if not process_list:
            return set()
        if isinstance(process_list, str):
            process_list = shlex.split(process_list)
        # process names from /proc only contain 15 characters
        return {p[0:15] for p in process_list}

    @staticmethod
    def iterate_children():
        """Iterates through all children process of the lutris client.
        This is not accurate since not all processes are started by
        lutris but are started by Systemd instead.
        """
        return Process(os.getpid()).iter_children()

    def iterate_processes(self):
        for child in self.iterate_children():
            if child.state == "Z":
                continue

            if child.name and child.name not in self.unmonitored_processes:
                yield child

    def is_alive(self, message=None):
        """Returns whether at least one watched process exists"""
        if message:
            sys.stdout.write("%s\n" % message)
        return next(self.iterate_processes(), None) is not None
